@@ -1,12 +1,28 @@
-/*	$OpenBSD: eval.c,v 1.33 2007/08/02 11:05:54 fgsch Exp $	*/
+/*	$OpenBSD: eval.c,v 1.35 2010/03/24 08:27:26 fgsch Exp $	*/
+
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
 
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/eval.c,v 1.47 2008/07/12 16:56:38 tg Exp $");
-
-#ifdef MKSH_SMALL
-#define MKSH_NOPWNAM
-#endif
+__RCSID("$MirOS: src/bin/mksh/eval.c,v 1.90 2010/07/17 22:09:33 tg Exp $");
 
 /*
  * string expansion
@@ -17,14 +33,14 @@ __RCSID("$MirOS: src/bin/mksh/eval.c,v 1.47 2008/07/12 16:56:38 tg Exp $");
 
 /* expansion generator state */
 typedef struct Expand {
-	/* int type; */		/* see expand() */
-	const char *str;	/* string */
+	/* int type; */			/* see expand() */
+	const char *str;		/* string */
 	union {
-		const char **strv;/* string[] */
-		struct shf *shf;  /* file */
-	} u;			/* source */
-	struct tbl *var;	/* variable in ${var..} */
-	short split;		/* split "$@" / call waitlast $() */
+		const char **strv;	/* string[] */
+		struct shf *shf;	/* file */
+	} u;				/* source */
+	struct tbl *var;		/* variable in ${var..} */
+	short split;			/* split "$@" / call waitlast $() */
 } Expand;
 
 #define	XBASE		0	/* scanning original */
@@ -51,6 +67,35 @@ static char *tilde(char *);
 static char *homedir(char *);
 #endif
 static void alt_expand(XPtrV *, char *, char *, char *, int);
+static size_t utflen(const char *);
+static void utfincptr(const char *, mksh_ari_t *);
+
+/* UTFMODE functions */
+static size_t
+utflen(const char *s)
+{
+	size_t n;
+
+	if (UTFMODE) {
+		n = 0;
+		while (*s) {
+			s += utf_ptradj(s);
+			++n;
+		}
+	} else
+		n = strlen(s);
+	return (n);
+}
+
+static void
+utfincptr(const char *s, mksh_ari_t *lp)
+{
+	const char *cp = s;
+
+	while ((*lp)--)
+		cp += utf_ptradj(cp);
+	*lp = cp - s;
+}
 
 /* compile and expand word */
 char *
@@ -66,7 +111,7 @@ substitute(const char *cp, int f)
 		internal_errorf("substitute");
 	source = sold;
 	afree(s, ATEMP);
-	return evalstr(yylval.cp, f);
+	return (evalstr(yylval.cp, f));
 }
 
 /*
@@ -88,7 +133,7 @@ eval(const char **ap, int f)
 	while (*ap != NULL)
 		expand(*ap++, &w, f);
 	XPput(w, NULL);
-	return (char **) XPclose(w) + 1;
+	return ((char **)XPclose(w) + 1);
 }
 
 /*
@@ -137,13 +182,14 @@ evalonestr(const char *cp, int f)
 
 /* for nested substitution: ${var:=$var2} */
 typedef struct SubType {
-	short	stype;		/* [=+-?%#] action after expanded word */
-	short	base;		/* begin position of expanded word */
-	short	f;		/* saved value of f (DOPAT, etc) */
-	short	quote;		/* saved value of quote (for ${..[%#]..}) */
 	struct tbl *var;	/* variable for ${var..} */
 	struct SubType *prev;	/* old type */
 	struct SubType *next;	/* poped type (to avoid re-allocating) */
+	short	stype;		/* [=+-?%#] action after expanded word */
+	short	base;		/* begin position of expanded word */
+	short	f;		/* saved value of f (DOPAT, etc) */
+	uint8_t	quotep;		/* saved value of quote (for ${..[%#]..}) */
+	uint8_t	quotew;		/* saved value of quote (for ${..[+-=]..}) */
 } SubType;
 
 void
@@ -191,7 +237,8 @@ expand(const char *cp,	/* input word */
 	doblank = 0;
 	make_magic = 0;
 	word = (f&DOBLANK) ? IFS_WS : IFS_WORD;
-	st_head.next = NULL;
+	/* clang doesn't know OSUBST comes before CSUBST */
+	memset(&st_head, 0, sizeof(st_head));
 	st = &st_head;
 
 	while (1) {
@@ -217,7 +264,7 @@ expand(const char *cp,	/* input word */
 				quote = 1;
 				continue;
 			case CQUOTE:
-				quote = 0;
+				quote = st->quotew;
 				continue;
 			case COMSUB:
 				tilde_ok = 0;
@@ -263,13 +310,12 @@ expand(const char *cp,	/* input word */
 					}
 				}
 				continue;
-			case OSUBST: /* ${{#}var{:}[=+-?#%]word} */
-			  /* format is:
-			   *   OSUBST [{x] plain-variable-part \0
-			   *     compiled-word-part CSUBST [}x]
-			   * This is where all syntax checking gets done...
-			   */
-			    {
+			case OSUBST: {	/* ${{#}var{:}[=+-?#%]word} */
+			/* format is:
+			 *	OSUBST [{x] plain-variable-part \0
+			 *	    compiled-word-part CSUBST [}x]
+			 * This is where all syntax checking gets done...
+			 */
 				const char *varname = ++sp; /* skip the { or x (}) */
 				int stype;
 				int slen = 0;
@@ -297,8 +343,7 @@ expand(const char *cp,	/* input word */
 					if (!st->next) {
 						SubType *newst;
 
-						newst = (SubType *)alloc(
-						    sizeof (SubType), ATEMP);
+						newst = alloc(sizeof(SubType), ATEMP);
 						newst->next = NULL;
 						newst->prev = st;
 						st->next = newst;
@@ -308,19 +353,15 @@ expand(const char *cp,	/* input word */
 					st->base = Xsavepos(ds, dp);
 					st->f = f;
 					st->var = x.var;
-					st->quote = quote;
+					st->quotew = st->quotep = quote;
 					/* skip qualifier(s) */
 					if (stype)
 						sp += slen;
 					switch (stype & 0x7f) {
 					case '0': {
 						char *beg, *mid, *end, *stg;
-						long from = 0, num = -1, flen;
+						mksh_ari_t from = 0, num = -1, flen, finc = 0;
 
-						/* ! DOBLANK,DOBRACE_,DOTILDE */
-						f = DOPAT | (f&DONTRUNCOMMAND) |
-						    DOTEMP_;
-						quote = 0;
 						beg = wdcopy(sp, ATEMP);
 						mid = beg + (wdscan(sp, ADELIM) - sp);
 						stg = beg + (wdscan(sp, CSUBST) - sp);
@@ -342,22 +383,27 @@ expand(const char *cp,	/* input word */
 						    &from, KSH_UNWIND_ERROR, true);
 						afree(stg, ATEMP);
 						if (end) {
-							evaluate(stg = wdstrip(mid, false, false),
+							evaluate(substitute(stg = wdstrip(mid, false, false), 0),
 							    &num, KSH_UNWIND_ERROR, true);
 							afree(stg, ATEMP);
 						}
 						afree(beg, ATEMP);
 						beg = str_val(st->var);
-						flen = strlen(beg);
+						flen = utflen(beg);
 						if (from < 0) {
 							if (-from < flen)
-								beg += flen + from;
+								finc = flen + from;
 						} else
-							beg += from < flen ? from : flen;
-						flen = strlen(beg);
+							finc = from < flen ? from : flen;
+						if (UTFMODE)
+							utfincptr(beg, &finc);
+						beg += finc;
+						flen = utflen(beg);
 						if (num < 0 || num > flen)
 							num = flen;
-						x.str = str_nsave(beg, num, ATEMP);
+						if (UTFMODE)
+							utfincptr(beg, &num);
+						strndupx(x.str, beg, num, ATEMP);
 						goto do_CSUBST;
 					}
 					case '/': {
@@ -365,24 +411,9 @@ expand(const char *cp,	/* input word */
 						char *pat, *rrep;
 						char *tpat0, *tpat1, *tpat2;
 
-						/* ! DOBLANK,DOBRACE_,DOTILDE */
-						f = DOPAT | (f&DONTRUNCOMMAND) |
-						    DOTEMP_;
-						quote = 0;
-
 						s = wdcopy(sp, ATEMP);
 						p = s + (wdscan(sp, ADELIM) - sp);
 						d = s + (wdscan(sp, CSUBST) - sp);
-#if 0
-						fprintf(stderr,
-						    "D: s=%p 〈%s〉\n"
-						    "   p=%p 〈%s〉\n"
-						    "   d=%p 〈%s〉\n",
-						    s, wdstrip(s, true, false),
-						    p, wdstrip(p, true, false),
-						    d, wdstrip(d, true, false));
-						fflush(stderr);
-#endif
 						if (p >= d)
 							goto unwind_substsyn;
 						p[-2] = EOS;
@@ -412,12 +443,6 @@ expand(const char *cp,	/* input word */
 							else
 								s++;
 						*d = '\0';
-#if 0
-						fprintf(stderr,
-						    "D: 〔%s｜%s〕→〔%s〕\n",
-						    tpat0, pat, rrep);
-						fflush(stderr);
-#endif
 						afree(tpat0, ATEMP);
 
 						/* reject empty pattern */
@@ -425,8 +450,8 @@ expand(const char *cp,	/* input word */
 							goto no_repl;
 
 						/* prepare string on which to work */
-						tpat0 = str_val(st->var);
-						sbeg = s = str_save(tpat0, ATEMP);
+						strdupx(s, str_val(st->var), ATEMP);
+						sbeg = s;
 
 						/* first see if we have any match at all */
 						tpat0 = pat;
@@ -444,12 +469,6 @@ expand(const char *cp,	/* input word */
 							tpat2 = tpat1 + 2;
 						}
  again_repl:
-#if 0
-						fprintf(stderr,
-						    "D: 「%s」 ← 〔%s｜%s〕\n",
-						    s, tpat0, rrep);
-						fflush(stderr);
-#endif
 						/* this would not be necessary if gmatchx would return
 						 * the start and end values of a match found, like re*
 						 */
@@ -477,7 +496,7 @@ expand(const char *cp,	/* input word */
 									break;
 								p--;
 							}
-						end = str_nsave(s, sbeg - s, ATEMP);
+						strndupx(end, s, sbeg - s, ATEMP);
 						d = shf_smprintf("%s%s%s", end, rrep, p);
 						afree(end, ATEMP);
 						sbeg = d + (sbeg - s) + strlen(rrep);
@@ -499,7 +518,7 @@ expand(const char *cp,	/* input word */
 						/* ! DOBLANK,DOBRACE_,DOTILDE */
 						f = DOPAT | (f&DONTRUNCOMMAND) |
 						    DOTEMP_;
-						quote = 0;
+						st->quotew = quote = 0;
 						/* Prepend open pattern (so |
 						 * in a trim will work as
 						 * expected)
@@ -512,7 +531,7 @@ expand(const char *cp,	/* input word */
 						 * after :s here is
 						 * non-standard ksh, but is
 						 * consistent with rules for
-						 * other assignments.  Not
+						 * other assignments. Not
 						 * sure what POSIX thinks of
 						 * this.
 						 * Not doing tilde expansion
@@ -533,7 +552,7 @@ expand(const char *cp,	/* input word */
 					case '?':
 						f &= ~DOBLANK;
 						f |= DOTEMP_;
-						/* FALLTHRU */
+						/* FALLTHROUGH */
 					default:
 						/* Enable tilde expansion */
 						tilde_ok = 1;
@@ -543,13 +562,13 @@ expand(const char *cp,	/* input word */
 					/* skip word */
 					sp += wdscan(sp, CSUBST) - sp;
 				continue;
-			    }
+			}
 			case CSUBST: /* only get here if expanding word */
  do_CSUBST:
 				sp++; /* ({) skip the } or x */
 				tilde_ok = 0;	/* in case of ${unset:-} */
 				*dp = '\0';
-				quote = st->quote;
+				quote = st->quotep;
 				f = st->f;
 				if (f&DOBLANK)
 					doblank--;
@@ -565,7 +584,10 @@ expand(const char *cp,	/* input word */
 					 */
 					x.str = trimsub(str_val(st->var),
 						dp, st->stype);
-					type = XSUB;
+					if (x.str[0] != '\0' || st->quotep)
+						type = XSUB;
+					else
+						type = XNULLSUB;
 					if (f&DOBLANK)
 						doblank++;
 					st = st->prev;
@@ -582,7 +604,7 @@ expand(const char *cp,	/* input word */
 					 * like x[i+=1] to be evaluated twice.
 					 */
 					/* Note: not exported by FEXPORT
-					 * in at&t ksh.
+					 * in AT&T ksh.
 					 */
 					/* XXX POSIX says readonly is only
 					 * fatal for special builtins (setstr
@@ -590,7 +612,7 @@ expand(const char *cp,	/* input word */
 					 */
 					len = strlen(dp) + 1;
 					setstr(st->var,
-					    debunk((char *)alloc(len, ATEMP),
+					    debunk(alloc(len, ATEMP),
 					    dp, len), KSH_UNWIND_ERROR);
 					x.str = str_val(st->var);
 					type = XSUB;
@@ -598,15 +620,14 @@ expand(const char *cp,	/* input word */
 						doblank++;
 					st = st->prev;
 					continue;
-				case '?':
-				    {
+				case '?': {
 					char *s = Xrestpos(ds, dp, st->base);
 
 					errorf("%s: %s", st->var->name,
 					    dp == s ?
 					    "parameter null or not set" :
 					    (debunk(s, s, strlen(s) + 1), s));
-				    }
+				}
 				case '0':
 				case '/':
 					dp = Xrestpos(ds, dp, st->base);
@@ -702,8 +723,9 @@ expand(const char *cp,	/* input word */
 				--newlines;
 			} else {
 				while ((c = shf_getc(x.u.shf)) == 0 || c == '\n')
-				    if (c == '\n')
-					    newlines++;	/* Save newlines */
+					if (c == '\n')
+						/* Save newlines */
+						newlines++;
 				if (newlines && c != EOF) {
 					shf_ungetc(c, x.u.shf);
 					c = '\n';
@@ -727,14 +749,14 @@ expand(const char *cp,	/* input word */
 		if (c == 0 || (!quote && (f & DOBLANK) && doblank &&
 		    !make_magic && ctype(c, C_IFS))) {
 			/* How words are broken up:
-			 *		   |       value of c
-			 *	  word	   |	ws	nws	0
+			 *			|	value of c
+			 *	word		|	ws	nws	0
 			 *	-----------------------------------
-			 *	IFS_WORD	w/WS	w/NWS	w
-			 *	IFS_WS		-/WS	w/NWS	-
-			 *	IFS_NWS		-/NWS	w/NWS	w
-			 *   (w means generate a word)
-			 * Note that IFS_NWS/0 generates a word (at&t ksh
+			 *	IFS_WORD		w/WS	w/NWS	w
+			 *	IFS_WS			-/WS	w/NWS	-
+			 *	IFS_NWS			-/NWS	w/NWS	w
+			 * (w means generate a word)
+			 * Note that IFS_NWS/0 generates a word (AT&T ksh
 			 * doesn't do this, but POSIX does).
 			 */
 			if (word == IFS_WORD ||
@@ -883,44 +905,79 @@ varsub(Expand *xp, const char *sp, const char *word,
 	int slen;
 	const char *p;
 	struct tbl *vp;
+	bool zero_ok = false;
 
-	if (sp[0] == '\0')	/* Bad variable name */
-		return -1;
+	if ((stype = sp[0]) == '\0')	/* Bad variable name */
+		return (-1);
 
 	xp->var = NULL;
 
-	/* ${#var}, string length or array size */
-	if (sp[0] == '#' && (c = sp[1]) != '\0') {
-		int zero_ok = 0;
-
-		/* Can't have any modifiers for ${#...} */
+	/*-
+	 * ${#var}, string length (-U: characters, +U: octets) or array size
+	 * ${%var}, string width (-U: screen columns, +U: octets)
+	 */
+	c = sp[1];
+	if (stype == '%' && c == '\0')
+		return (-1);
+	if ((stype == '#' || stype == '%') && c != '\0') {
+		/* Can't have any modifiers for ${#...} or ${%...} */
 		if (*word != CSUBST)
-			return -1;
+			return (-1);
 		sp++;
 		/* Check for size of array */
 		if ((p = cstrchr(sp, '[')) && (p[1] == '*' || p[1] == '@') &&
 		    p[2] == ']') {
 			int n = 0;
 
+			if (stype != '#')
+				return (-1);
 			vp = global(arrayname(sp));
 			if (vp->flag & (ISSET|ARRAY))
-				zero_ok = 1;
+				zero_ok = true;
 			for (; vp; vp = vp->u.array)
 				if (vp->flag & ISSET)
 					n++;
 			c = n;
-		} else if (c == '*' || c == '@')
+		} else if (c == '*' || c == '@') {
+			if (stype != '#')
+				return (-1);
 			c = e->loc->argc;
-		else {
+		} else {
 			p = str_val(global(sp));
 			zero_ok = p != null;
-			c = strlen(p);
+			if (stype == '#')
+				c = utflen(p);
+			else {
+				/* partial utf_mbswidth reimplementation */
+				const char *s = p;
+				unsigned int wc;
+				size_t len;
+				int cw;
+
+				c = 0;
+				while (*s) {
+					if (!UTFMODE || (len = utf_mbtowc(&wc,
+					    s)) == (size_t)-1)
+						/* not UTFMODE or not UTF-8 */
+						wc = (unsigned char)(*s++);
+					else
+						/* UTFMODE and UTF-8 */
+						s += len;
+					/* wc == char or wchar at s++ */
+					if ((cw = utf_wcwidth(wc)) == -1) {
+						/* 646, 8859-1, 10646 C0/C1 */
+						c = -1;
+						break;
+					}
+					c += cw;
+				}
+			}
 		}
 		if (Flag(FNOUNSET) && c == 0 && !zero_ok)
 			errorf("%s: parameter not set", sp);
 		*stypep = 0; /* unqualified variable/string substitution */
-		xp->str = shf_smprintf("%lu", (unsigned long)c);
-		return XSUB;
+		xp->str = shf_smprintf("%d", c);
+		return (XSUB);
 	}
 
 	/* Check for qualifiers in word part */
@@ -951,9 +1008,9 @@ varsub(Expand *xp, const char *sp, const char *word,
 			slen += 2;
 		}
 	} else if (stype)	/* : is not ok */
-		return -1;
+		return (-1);
 	if (!stype && *word != CSUBST)
-		return -1;
+		return (-1);
 	*stypep = stype;
 	*slenp = slen;
 
@@ -963,18 +1020,21 @@ varsub(Expand *xp, const char *sp, const char *word,
 		case '=':	/* can't assign to a vector */
 		case '%':	/* can't trim a vector (yet) */
 		case '#':
-			return -1;
+		case '0':
+		case '/':
+			return (-1);
 		}
 		if (e->loc->argc == 0) {
 			xp->str = null;
 			xp->var = global(sp);
 			state = c == '@' ? XNULLSUB : XSUB;
 		} else {
-			xp->u.strv = (const char **) e->loc->argv + 1;
+			xp->u.strv = (const char **)e->loc->argv + 1;
 			xp->str = *xp->u.strv++;
 			xp->split = c == '@'; /* $@ */
 			state = XARG;
 		}
+		zero_ok = true;	/* POSIX 2009? */
 	} else {
 		if ((p = cstrchr(sp, '[')) && (p[1] == '*' || p[1] == '@') &&
 		    p[2] == ']') {
@@ -985,14 +1045,20 @@ varsub(Expand *xp, const char *sp, const char *word,
 			case '%':	/* can't trim a vector (yet) */
 			case '#':
 			case '?':
-				return -1;
+			case '0':
+			case '/':
+				return (-1);
 			}
 			XPinit(wv, 32);
+			if ((c = sp[0]) == '!')
+				++sp;
 			vp = global(arrayname(sp));
 			for (; vp; vp = vp->u.array) {
 				if (!(vp->flag&ISSET))
 					continue;
-				XPput(wv, str_val(vp));
+				XPput(wv, c == '!' ? shf_smprintf("%lu",
+				    arrayindex(vp)) :
+				    str_val(vp));
 			}
 			if (XPsize(wv) == 0) {
 				xp->str = null;
@@ -1000,7 +1066,7 @@ varsub(Expand *xp, const char *sp, const char *word,
 				XPfree(wv);
 			} else {
 				XPput(wv, 0);
-				xp->u.strv = (const char **) XPptrv(wv);
+				xp->u.strv = (const char **)XPptrv(wv);
 				xp->str = *xp->u.strv++;
 				xp->split = p[1] == '@'; /* ${foo[@]} */
 				state = XARG;
@@ -1009,9 +1075,24 @@ varsub(Expand *xp, const char *sp, const char *word,
 			/* Can't assign things like $! or $1 */
 			if ((stype & 0x7f) == '=' &&
 			    ctype(*sp, C_VAR1 | C_DIGIT))
-				return -1;
-			xp->var = global(sp);
-			xp->str = str_val(xp->var);
+				return (-1);
+			if (*sp == '!' && sp[1]) {
+				++sp;
+				xp->var = global(sp);
+				if (cstrchr(sp, '[')) {
+					if (xp->var->flag & ISSET)
+						xp->str = shf_smprintf("%lu",
+						    arrayindex(xp->var));
+					else
+						xp->str = null;
+				} else if (xp->var->flag & ISSET)
+					xp->str = xp->var->name;
+				else
+					xp->str = "0";	/* ksh93 compat */
+			} else {
+				xp->var = global(sp);
+				xp->str = str_val(xp->var);
+			}
 			state = XSUB;
 		}
 	}
@@ -1022,10 +1103,10 @@ varsub(Expand *xp, const char *sp, const char *word,
 	    (((stype&0x80) ? *xp->str=='\0' : xp->str==null) ? /* undef? */
 	    c == '=' || c == '-' || c == '?' : c == '+'))
 		state = XBASE;	/* expand word instead of variable value */
-	if (Flag(FNOUNSET) && xp->str == null &&
+	if (Flag(FNOUNSET) && xp->str == null && !zero_ok &&
 	    (ctype(c, C_SUBOP2) || (state != XBASE && c != '+')))
 		errorf("%s: parameter not set", sp);
-	return state;
+	return (state);
 }
 
 /*
@@ -1046,7 +1127,7 @@ comsub(Expand *xp, const char *cp)
 	source = sold;
 
 	if (t == NULL)
-		return XBASE;
+		return (XBASE);
 
 	if (t != NULL && t->type == TCOM && /* $(<file) */
 	    *t->args == NULL && *t->vars == NULL && t->ioact != NULL) {
@@ -1070,14 +1151,14 @@ comsub(Expand *xp, const char *cp)
 			ksh_dup2(pv[1], 1, false);
 			close(pv[1]);
 		}
-		execute(t, XFORK|XXCOM|XPIPEO);
+		execute(t, XFORK|XXCOM|XPIPEO, NULL);
 		restfd(1, ofd1);
 		startlast();
 		xp->split = 1;	/* waitlast() */
 	}
 
 	xp->u.shf = shf;
-	return XCOM;
+	return (XCOM);
 }
 
 /*
@@ -1092,11 +1173,11 @@ trimsub(char *str, char *pat, int how)
 
 	switch (how & 0xFF) {
 	case '#':		/* shortest at beginning */
-		for (p = str; p <= end; p++) {
+		for (p = str; p <= end; p += utf_ptradj(p)) {
 			c = *p; *p = '\0';
 			if (gmatchx(str, pat, false)) {
 				*p = c;
-				return p;
+				return (p);
 			}
 			*p = c;
 		}
@@ -1106,26 +1187,37 @@ trimsub(char *str, char *pat, int how)
 			c = *p; *p = '\0';
 			if (gmatchx(str, pat, false)) {
 				*p = c;
-				return p;
+				return (p);
 			}
 			*p = c;
 		}
 		break;
 	case '%':		/* shortest match at end */
-		for (p = end; p >= str; p--) {
+		p = end;
+		while (p >= str) {
 			if (gmatchx(p, pat, false))
-				return str_nsave(str, p - str, ATEMP);
+				goto trimsub_match;
+			if (UTFMODE) {
+				char *op = p;
+				while ((p-- > str) && ((*p & 0xC0) == 0x80))
+					;
+				if ((p < str) || (p + utf_ptradj(p) != op))
+					p = op - 1;
+			} else
+				--p;
 		}
 		break;
 	case '%'|0x80:		/* longest match at end */
-		for (p = str; p <= end; p++) {
-			if (gmatchx(p, pat, false))
-				return str_nsave(str, p - str, ATEMP);
-		}
+		for (p = str; p <= end; p++)
+			if (gmatchx(p, pat, false)) {
+ trimsub_match:
+				strndupx(end, str, p - str, ATEMP);
+				return (end);
+			}
 		break;
 	}
 
-	return str;		/* no match, return string */
+	return (str);		/* no match, return string */
 }
 
 /*
@@ -1133,7 +1225,7 @@ trimsub(char *str, char *pat, int how)
  * Name derived from V6's /etc/glob, the program that expanded filenames.
  */
 
-/* XXX cp not const 'cause slashes are temporarily replaced with nulls... */
+/* XXX cp not const 'cause slashes are temporarily replaced with NULs... */
 static void
 glob(char *cp, XPtrV *wp, int markdirs)
 {
@@ -1143,7 +1235,7 @@ glob(char *cp, XPtrV *wp, int markdirs)
 		XPput(*wp, debunk(cp, cp, strlen(cp) + 1));
 	else
 		qsort(XPptrv(*wp) + oldsize, XPsize(*wp) - oldsize,
-		    sizeof (void *), xstrcmp);
+		    sizeof(void *), xstrcmp);
 }
 
 #define GF_NONE		0
@@ -1151,7 +1243,7 @@ glob(char *cp, XPtrV *wp, int markdirs)
 #define GF_GLOBBED	BIT(1)		/* some globbing has been done */
 #define GF_MARKDIR	BIT(2)		/* add trailing / to directories */
 
-/* Apply file globbing to cp and store the matching files in wp.  Returns
+/* Apply file globbing to cp and store the matching files in wp. Returns
  * the number of matches found.
  */
 int
@@ -1165,7 +1257,7 @@ glob_str(char *cp, XPtrV *wp, int markdirs)
 	globit(&xs, &xp, cp, wp, markdirs ? GF_MARKDIR : GF_NONE);
 	Xfree(xs, xp);
 
-	return XPsize(*wp) - oldsize;
+	return (XPsize(*wp) - oldsize);
 }
 
 static void
@@ -1222,7 +1314,8 @@ globit(XString *xs,	/* dest string */
 				*xp = '\0';
 			}
 		}
-		XPput(*wp, str_nsave(Xstring(*xs, xp), Xlength(*xs, xp), ATEMP));
+		strndupx(np, Xstring(*xs, xp), Xlength(*xs, xp), ATEMP);
+		XPput(*wp, np);
 		return;
 	}
 
@@ -1262,7 +1355,7 @@ globit(XString *xs,	/* dest string */
 		int len;
 		int prefix_len;
 
-		/* xp = *xpp;	   copy_non_glob() may have re-alloc'd xs */
+		/* xp = *xpp;	copy_non_glob() may have re-alloc'd xs */
 		*xp = '\0';
 		prefix_len = Xlength(*xs, xp);
 		dirp = opendir(prefix_len ? Xstring(*xs, xp) : ".");
@@ -1304,8 +1397,8 @@ debunk(char *dp, const char *sp, size_t dlen)
 
 	if ((s = cstrchr(sp, MAGIC))) {
 		if (s - sp >= (ssize_t)dlen)
-			return dp;
-		memcpy(dp, sp, s - sp);
+			return (dp);
+		memmove(dp, sp, s - sp);
 		for (d = dp + (s - sp); *s && (d - dp < (ssize_t)dlen); s++)
 			if (!ISMAGIC(*s) || !(*++s & 0x80) ||
 			    !vstrchr("*+?@! ", *s & 0x7f))
@@ -1320,10 +1413,10 @@ debunk(char *dp, const char *sp, size_t dlen)
 		*d = '\0';
 	} else if (dp != sp)
 		strlcpy(dp, sp, dlen);
-	return dp;
+	return (dp);
 }
 
-/* Check if p is an unquoted name, possibly followed by a / or :.  If so
+/* Check if p is an unquoted name, possibly followed by a / or :. If so
  * puts the expanded version in *dcp,dp and returns a pointer in p just
  * past the name, otherwise returns 0.
  */
@@ -1357,7 +1450,7 @@ maybe_expand_tilde(const char *p, XString *dsp, char **dpp, int isassign)
 		*dpp = dp;
 		r = p;
 	}
-	return r;
+	return (r);
 }
 
 /*
@@ -1403,11 +1496,11 @@ homedir(char *name)
 
 		pw = getpwnam(name);
 		if (pw == NULL)
-			return NULL;
-		ap->val.s = str_save(pw->pw_dir, APERM);
+			return (NULL);
+		strdupx(ap->val.s, pw->pw_dir, APERM);
 		ap->flag |= DEFINED|ISSET|ALLOC;
 	}
-	return ap->val.s;
+	return (ap->val.s);
 }
 #endif
 
@@ -1442,7 +1535,7 @@ alt_expand(XPtrV *wp, char *start, char *exp_start, char *end, int fdo)
 	/* no valid expansions... */
 	if (!p || count != 0) {
 		/* Note that given a{{b,c} we do not expand anything (this is
-		 * what at&t ksh does.  This may be changed to do the {b,c}
+		 * what AT&T ksh does. This may be changed to do the {b,c}
 		 * expansion. }
 		 */
 		if (fdo & DOGLOB)
@@ -1466,19 +1559,19 @@ alt_expand(XPtrV *wp, char *start, char *exp_start, char *end, int fdo)
 				count++;
 			else if ((*p == CBRACE && --count == 0) ||
 			    (*p == ',' && count == 1)) {
-				char *new;
+				char *news;
 				int l1, l2, l3;
 
 				l1 = brace_start - start;
 				l2 = (p - 1) - field_start;
 				l3 = end - brace_end;
-				new = (char *)alloc(l1 + l2 + l3 + 1, ATEMP);
-				memcpy(new, start, l1);
-				memcpy(new + l1, field_start, l2);
-				memcpy(new + l1 + l2, brace_end, l3);
-				new[l1 + l2 + l3] = '\0';
-				alt_expand(wp, new, new + l1,
-				    new + l1 + l2 + l3, fdo);
+				news = alloc(l1 + l2 + l3 + 1, ATEMP);
+				memcpy(news, start, l1);
+				memcpy(news + l1, field_start, l2);
+				memcpy(news + l1 + l2, brace_end, l3);
+				news[l1 + l2 + l3] = '\0';
+				alt_expand(wp, news, news + l1,
+				    news + l1 + l2 + l3, fdo);
 				field_start = p + 1;
 			}
 		}
