@@ -1,18 +1,41 @@
-/*	$OpenBSD: exec.c,v 1.48 2007/09/05 19:02:01 otto Exp $	*/
+/*	$OpenBSD: exec.c,v 1.49 2009/01/29 23:27:26 jaredy Exp $	*/
+
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
 
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/exec.c,v 1.44 2008/05/17 18:46:58 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/exec.c,v 1.75 2010/07/17 22:09:34 tg Exp $");
+
+#ifndef MKSH_DEFAULT_EXECSHELL
+#define MKSH_DEFAULT_EXECSHELL	"/bin/sh"
+#endif
 
 static int comexec(struct op *, struct tbl *volatile, const char **,
-    int volatile);
-static void scriptexec(struct op *, const char **)
-    __attribute__((noreturn));
+    int volatile, volatile int *);
+static void scriptexec(struct op *, const char **) MKSH_A_NORETURN;
 static int call_builtin(struct tbl *, const char **);
 static int iosetup(struct ioword *, struct tbl *);
 static int herein(const char *, int);
 static const char *do_selectargs(const char **, bool);
-static int dbteste_isa(Test_env *, Test_meta);
+static Test_op dbteste_isa(Test_env *, Test_meta);
 static const char *dbteste_getopnd(Test_env *, Test_op, bool);
 static void dbteste_error(Test_env *, int, const char *);
 
@@ -21,10 +44,11 @@ static void dbteste_error(Test_env *, int, const char *);
  */
 int
 execute(struct op *volatile t,
-    volatile int flags)		/* if XEXEC don't fork */
+    volatile int flags,		/* if XEXEC don't fork */
+    volatile int * volatile xerrok)
 {
 	int i;
-	volatile int rv = 0;
+	volatile int rv = 0, dummy = 0;
 	int pv[2];
 	const char ** volatile ap;
 	char ** volatile up;
@@ -33,17 +57,22 @@ execute(struct op *volatile t,
 	struct tbl *tp = NULL;
 
 	if (t == NULL)
-		return 0;
+		return (0);
+
+	/* Caller doesn't care if XERROK should propagate. */
+	if (xerrok == NULL)
+		xerrok = &dummy;
 
 	if ((flags&XFORK) && !(flags&XEXEC) && t->type != TPIPE)
-		return exchild(t, flags & ~XTIME, -1); /* run in sub-process */
+		/* run in sub-process */
+		return (exchild(t, flags & ~XTIME, xerrok, -1));
 
 	newenv(E_EXEC);
 	if (trap)
 		runtraps(0);
 
 	if (t->type == TCOM) {
-		/* Clear subst_exstat before argument expansion.  Used by
+		/* Clear subst_exstat before argument expansion. Used by
 		 * null commands (see comexec() and c_eval()) and by c_set().
 		 */
 		subst_exstat = 0;
@@ -72,9 +101,9 @@ execute(struct op *volatile t,
 	flags &= ~XTIME;
 
 	if (t->ioact != NULL || t->type == TPIPE || t->type == TCOPROC) {
-		e->savefd = (short *)alloc(sizeofN(short, NUFILE), ATEMP);
+		e->savefd = alloc(NUFILE * sizeof(short), ATEMP);
 		/* initialise to not redirected */
-		memset(e->savefd, 0, sizeofN(short, NUFILE));
+		memset(e->savefd, 0, NUFILE * sizeof(short));
 	}
 
 	/* do redirection, to be restored in quitenv() */
@@ -95,11 +124,11 @@ execute(struct op *volatile t,
 
 	switch (t->type) {
 	case TCOM:
-		rv = comexec(t, tp, (const char **)ap, flags);
+		rv = comexec(t, tp, (const char **)ap, flags, xerrok);
 		break;
 
 	case TPAREN:
-		rv = execute(t->left, flags|XFORK);
+		rv = execute(t->left, flags | XFORK, xerrok);
 		break;
 
 	case TPIPE:
@@ -110,12 +139,14 @@ execute(struct op *volatile t,
 		while (t->type == TPIPE) {
 			openpipe(pv);
 			ksh_dup2(pv[1], 1, false); /* stdout of curr */
-			/* Let exchild() close pv[0] in child
+			/**
+			 * Let exchild() close pv[0] in child
 			 * (if this isn't done, commands like
-			 *    (: ; cat /etc/termcap) | sleep 1
-			 *  will hang forever).
+			 *	(: ; cat /etc/termcap) | sleep 1
+			 * will hang forever).
 			 */
-			exchild(t->left, flags|XPIPEO|XCCLOSE, pv[0]);
+			exchild(t->left, flags | XPIPEO | XCCLOSE,
+			    NULL, pv[0]);
 			ksh_dup2(pv[0], 0, false); /* stdin of next */
 			closepipe(pv);
 			flags |= XPIPEI;
@@ -124,22 +155,21 @@ execute(struct op *volatile t,
 		restfd(1, e->savefd[1]); /* stdout of last */
 		e->savefd[1] = 0; /* no need to re-restore this */
 		/* Let exchild() close 0 in parent, after fork, before wait */
-		i = exchild(t, flags|XPCLOSE, 0);
+		i = exchild(t, flags | XPCLOSE, xerrok, 0);
 		if (!(flags&XBGND) && !(flags&XXCOM))
 			rv = i;
 		break;
 
 	case TLIST:
 		while (t->type == TLIST) {
-			execute(t->left, flags & XERROK);
+			execute(t->left, flags & XERROK, NULL);
 			t = t->right;
 		}
-		rv = execute(t, flags & XERROK);
+		rv = execute(t, flags & XERROK, xerrok);
 		break;
 
-	case TCOPROC:
-	    {
-		sigset_t	omask;
+	case TCOPROC: {
+		sigset_t omask;
 
 		/* Block sigchild as we are using things changed in the
 		 * signal handler
@@ -191,34 +221,37 @@ execute(struct op *volatile t,
 		 * job is actually created.
 		 */
 		flags &= ~XEXEC;
-		exchild(t->left, flags|XBGND|XFORK|XCOPROC|XCCLOSE,
-		    coproc.readw);
+		exchild(t->left, flags | XBGND | XFORK | XCOPROC | XCCLOSE,
+		    NULL, coproc.readw);
 		break;
-	    }
+	}
 
 	case TASYNC:
 		/* XXX non-optimal, I think - "(foo &)", forks for (),
-		 * forks again for async...  parent should optimise
+		 * forks again for async... parent should optimise
 		 * this to "foo &"...
 		 */
-		rv = execute(t->left, (flags&~XEXEC)|XBGND|XFORK);
+		rv = execute(t->left, (flags&~XEXEC)|XBGND|XFORK, xerrok);
 		break;
 
 	case TOR:
 	case TAND:
-		rv = execute(t->left, XERROK);
-		if (t->right != NULL && (rv == 0) == (t->type == TAND))
-			rv = execute(t->right, flags & XERROK);
-		else
-			flags |= XERROK;
+		rv = execute(t->left, XERROK, xerrok);
+		if ((rv == 0) == (t->type == TAND))
+			rv = execute(t->right, XERROK, xerrok);
+		flags |= XERROK;
+		if (xerrok)
+			*xerrok = 1;
 		break;
 
 	case TBANG:
-		rv = !execute(t->right, XERROK);
+		rv = !execute(t->right, XERROK, xerrok);
+		flags |= XERROK;
+		if (xerrok)
+			*xerrok = 1;
 		break;
 
-	case TDBRACKET:
-	    {
+	case TDBRACKET: {
 		Test_env te;
 
 		te.flags = TEF_DBRACKET;
@@ -230,11 +263,10 @@ execute(struct op *volatile t,
 
 		rv = test_parse(&te);
 		break;
-	    }
+	}
 
 	case TFOR:
-	case TSELECT:
-	    {
+	case TSELECT: {
 		volatile bool is_first = true;
 		ap = (t->vars == NULL) ? e->loc->argv + 1 :
 		    (const char **)eval((const char **)t->vars,
@@ -257,7 +289,7 @@ execute(struct op *volatile t,
 		if (t->type == TFOR) {
 			while (*ap != NULL) {
 				setstr(global(t->str), *ap++, KSH_UNWIND_ERROR);
-				rv = execute(t->left, flags & XERROK);
+				rv = execute(t->left, flags & XERROK, xerrok);
 			}
 		} else { /* TSELECT */
 			for (;;) {
@@ -267,11 +299,11 @@ execute(struct op *volatile t,
 				}
 				is_first = false;
 				setstr(global(t->str), cp, KSH_UNWIND_ERROR);
-				rv = execute(t->left, flags & XERROK);
+				execute(t->left, flags & XERROK, xerrok);
 			}
 		}
-	    }
 		break;
+	}
 
 	case TWHILE:
 	case TUNTIL:
@@ -290,17 +322,18 @@ execute(struct op *volatile t,
 			}
 		}
 		rv = 0; /* in case of a continue */
-		while ((execute(t->left, XERROK) == 0) == (t->type == TWHILE))
-			rv = execute(t->right, flags & XERROK);
+		while ((execute(t->left, XERROK, NULL) == 0) ==
+		    (t->type == TWHILE))
+			rv = execute(t->right, flags & XERROK, xerrok);
 		break;
 
 	case TIF:
 	case TELIF:
 		if (t->right == NULL)
 			break;	/* should be error */
-		rv = execute(t->left, XERROK) == 0 ?
-		    execute(t->right->left, flags & XERROK) :
-		    execute(t->right->right, flags & XERROK);
+		rv = execute(t->left, XERROK, NULL) == 0 ?
+		    execute(t->right->left, flags & XERROK, xerrok) :
+		    execute(t->right->right, flags & XERROK, xerrok);
 		break;
 
 	case TCASE:
@@ -312,11 +345,11 @@ execute(struct op *volatile t,
 				goto Found;
 		break;
  Found:
-		rv = execute(t->left, flags & XERROK);
+		rv = execute(t->left, flags & XERROK, xerrok);
 		break;
 
 	case TBRACE:
-		rv = execute(t->left, flags & XERROK);
+		rv = execute(t->left, flags & XERROK, xerrok);
 		break;
 
 	case TFUNCT:
@@ -327,7 +360,7 @@ execute(struct op *volatile t,
 		/* Clear XEXEC so nested execute() call doesn't exit
 		 * (allows "ls -l | time grep foo").
 		 */
-		rv = timex(t, flags & ~XEXEC);
+		rv = timex(t, flags & ~XEXEC, xerrok);
 		break;
 
 	case TEXEC:		/* an eval'd TCOM */
@@ -353,12 +386,13 @@ execute(struct op *volatile t,
 	quitenv(NULL);		/* restores IO */
 	if ((flags&XEXEC))
 		unwind(LEXIT);	/* exit child */
-	if (rv != 0 && !(flags & XERROK)) {
+	if (rv != 0 && !(flags & XERROK) &&
+	    (xerrok == NULL || !*xerrok)) {
 		trapsig(SIGERR_);
 		if (Flag(FERREXIT))
 			unwind(LERROR);
 	}
-	return rv;
+	return (rv);
 }
 
 /*
@@ -367,7 +401,7 @@ execute(struct op *volatile t,
 
 static int
 comexec(struct op *t, struct tbl *volatile tp, const char **ap,
-    volatile int flags)
+    volatile int flags, volatile int *xerrok)
 {
 	int i;
 	volatile int rv = 0;
@@ -377,9 +411,10 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 	int type_flags;
 	int keepasn_ok;
 	int fcflags = FC_BI|FC_FUNC|FC_PATH;
-	int bourne_function_call = 0;
+	bool bourne_function_call = false;
+	struct block *l_expand, *l_assign;
 
-	/* snag the last argument for $_ XXX not the same as at&t ksh,
+	/* snag the last argument for $_ XXX not the same as AT&T ksh,
 	 * which only seems to set $_ after a newline (but not in
 	 * functions/dot scripts, but in interactive and script) -
 	 * perhaps save last arg here and set it in shell()?.
@@ -393,16 +428,16 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 	}
 
 	/* Deal with the shell builtins builtin, exec and command since
-	 * they can be followed by other commands.  This must be done before
+	 * they can be followed by other commands. This must be done before
 	 * we know if we should create a local block which must be done
 	 * before we can do a path search (in case the assignments change
 	 * PATH).
 	 * Odd cases:
-	 *   FOO=bar exec >/dev/null		FOO is kept but not exported
-	 *   FOO=bar exec foobar		FOO is exported
-	 *   FOO=bar command exec >/dev/null	FOO is neither kept nor exported
-	 *   FOO=bar command			FOO is neither kept nor exported
-	 *   PATH=... foobar			use new PATH in foobar search
+	 *	FOO=bar exec >/dev/null		FOO is kept but not exported
+	 *	FOO=bar exec foobar		FOO is exported
+	 *	FOO=bar command exec >/dev/null	FOO is neither kept nor exported
+	 *	FOO=bar command			FOO is neither kept nor exported
+	 *	PATH=... foobar			use new PATH in foobar search
 	 */
 	keepasn_ok = 1;
 	while (tp && tp->type == CSHELL) {
@@ -457,6 +492,7 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 			break;
 		tp = findcom(ap[0], fcflags & (FC_BI|FC_FUNC));
 	}
+	l_expand = e->loc;
 	if (keepasn_ok && (!ap[0] || (tp && (tp->flag & KEEPASN))))
 		type_flags = 0;
 	else {
@@ -465,15 +501,21 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 		/* ksh functions don't keep assignments, POSIX functions do. */
 		if (keepasn_ok && tp && tp->type == CFUNC &&
 		    !(tp->flag & FKSH)) {
-			bourne_function_call = 1;
-			type_flags = 0;
+			bourne_function_call = true;
+			type_flags = EXPORT;
 		} else
 			type_flags = LOCAL|LOCAL_COPY|EXPORT;
 	}
+	l_assign = e->loc;
 	if (Flag(FEXPORT))
 		type_flags |= EXPORT;
 	for (i = 0; t->vars[i]; i++) {
+		/* do NOT lookup in the new var/fn block just created */
+		e->loc = l_expand;
 		cp = evalstr(t->vars[i], DOASNTILDE);
+		e->loc = l_assign;
+		/* but assign in there as usual */
+
 		if (Flag(FXTRACE)) {
 			if (i == 0)
 				shf_fprintf(shl_out, "%s",
@@ -505,9 +547,8 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 		rv = call_builtin(tp, (const char **)ap);
 		break;
 
-	case CFUNC:			/* function call */
-	    {
-		volatile char old_xflag;
+	case CFUNC: {			/* function call */
+		volatile unsigned char old_xflag;
 		volatile Tflag old_inuse;
 		const char *volatile old_kshname;
 
@@ -561,7 +602,7 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 			;
 		e->loc->argc = i - 1;
 		/* ksh-style functions handle getopts sanely,
-		 * bourne/posix functions are insane...
+		 * Bourne/POSIX functions are insane...
 		 */
 		if (tp->flag & FKSH) {
 			e->loc->flags |= BF_DOGETOPTS;
@@ -578,15 +619,15 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 		e->type = E_FUNC;
 		i = sigsetjmp(e->jbuf, 0);
 		if (i == 0) {
-			/* seems odd to pass XERROK here, but at&t ksh does */
-			exstat = execute(tp->val.t, flags & XERROK);
+			/* seems odd to pass XERROK here, but AT&T ksh does */
+			exstat = execute(tp->val.t, flags & XERROK, xerrok);
 			i = LRETURN;
 		}
 		kshname = old_kshname;
 		Flag(FXTRACE) = old_xflag;
 		tp->flag = (tp->flag & ~FINUSE) | old_inuse;
-		/* Were we deleted while executing?  If so, free the execution
-		 * tree.  todo: Unfortunately, the table entry is never re-used
+		/* Were we deleted while executing? If so, free the execution
+		 * tree. todo: Unfortunately, the table entry is never re-used
 		 * until the lookup table is expanded.
 		 */
 		if ((tp->flag & (FDELETE|FINUSE)) == FDELETE) {
@@ -613,14 +654,14 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 			internal_errorf("CFUNC %d", i);
 		}
 		break;
-	    }
+	}
 
 	case CEXEC:		/* executable command */
 	case CTALIAS:		/* tracked alias */
 		if (!(tp->flag&ISSET)) {
 			/* errno_ will be set if the named command was found
 			 * but could not be executed (permissions, no execute
-			 * bit, directory, etc).  Print out a (hopefully)
+			 * bit, directory, etc). Print out a (hopefully)
 			 * useful error message and set the exit status to 126.
 			 */
 			if (tp->u2.errno_) {
@@ -641,7 +682,11 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 
 		if (flags&XEXEC) {
 			j_exit();
-			if (!(flags&XBGND) || Flag(FMONITOR)) {
+			if (!(flags&XBGND)
+#ifndef MKSH_UNEMPLOYED
+			    || Flag(FMONITOR)
+#endif
+			    ) {
 				setexecsig(&sigtraps[SIGINT], SS_RESTORE_ORIG);
 				setexecsig(&sigtraps[SIGQUIT], SS_RESTORE_ORIG);
 			}
@@ -652,7 +697,7 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 		texec.left = t;	/* for tprint */
 		texec.str = tp->val.s;
 		texec.args = ap;
-		rv = exchild(&texec, flags, -1);
+		rv = exchild(&texec, flags, xerrok, -1);
 		break;
 	}
  Leave:
@@ -660,7 +705,7 @@ comexec(struct op *t, struct tbl *volatile tp, const char **ap,
 		exstat = rv;
 		unwind(LLEAVE);
 	}
-	return rv;
+	return (rv);
 }
 
 static void
@@ -678,27 +723,27 @@ scriptexec(struct op *tp, const char **ap)
 	if (sh && *sh)
 		sh = search(sh, path, X_OK, NULL);
 	if (!sh || !*sh)
-		sh = "/bin/sh";
+		sh = MKSH_DEFAULT_EXECSHELL;
 
 	*tp->args-- = tp->str;
 
 #ifndef MKSH_SMALL
 	if ((fd = open(tp->str, O_RDONLY)) >= 0) {
 		/* read first MAXINTERP octets from file */
-		if (read(fd, buf, sizeof (buf)) <= 0)
+		if (read(fd, buf, sizeof(buf)) <= 0)
 			/* read error -> no good */
 			buf[0] = '\0';
 		close(fd);
 		/* scan for newline (or CR) or NUL _before_ end of buffer */
 		cp = (unsigned char *)buf;
-		while ((char *)cp < (buf + sizeof (buf)))
+		while ((char *)cp < (buf + sizeof(buf)))
 			if (*cp == '\0' || *cp == '\n' || *cp == '\r') {
 				*cp = '\0';
 				break;
 			} else
 				++cp;
 		/* if the shebang line is longer than MAXINTERP, bail out */
-		if ((char *)cp >= (buf + sizeof (buf)))
+		if ((char *)cp >= (buf + sizeof(buf)))
 			goto noshebang;
 		/* skip UTF-8 Byte Order Mark, if present */
 		cp = (unsigned char *)buf;
@@ -739,7 +784,7 @@ scriptexec(struct op *tp, const char **ap)
 		    (fd == /* ECOFF_M68K */ 0x0150 || fd == 0x5001) ||
 		    (fd == /* ECOFF_SH */   0x0500 || fd == 0x0005) ||
 		    (fd == 0x7F45 && buf[2] == 'L' && buf[3] == 'F') ||
-		    (fd == /* “MZ” */ 0x4D5A) ||
+		    (fd == /* "MZ" */ 0x4D5A) ||
 		    (fd == /* gzip */ 0x1F8B))
 			errorf("%s: not executable: magic %04X", tp->str, fd);
 	}
@@ -766,11 +811,11 @@ shcomexec(const char **wp)
 }
 
 /*
- * Search function tables for a function.  If create set, a table entry
+ * Search function tables for a function. If create set, a table entry
  * is created if none is found.
  */
 struct tbl *
-findfunc(const char *name, unsigned int h, int create)
+findfunc(const char *name, uint32_t h, bool create)
 {
 	struct block *l;
 	struct tbl *tp = NULL;
@@ -787,24 +832,24 @@ findfunc(const char *name, unsigned int h, int create)
 			break;
 		}
 	}
-	return tp;
+	return (tp);
 }
 
 /*
- * define function.  Returns 1 if function is being undefined (t == 0) and
+ * define function. Returns 1 if function is being undefined (t == 0) and
  * function did not exist, returns 0 otherwise.
  */
 int
 define(const char *name, struct op *t)
 {
 	struct tbl *tp;
-	int was_set = 0;
+	bool was_set = false;
 
 	while (1) {
 		tp = findfunc(name, hash(name), true);
 
 		if (tp->flag & ISSET)
-			was_set = 1;
+			was_set = true;
 		/* If this function is currently being executed, we zap this
 		 * table entry so findfunc() won't see it
 		 */
@@ -823,7 +868,7 @@ define(const char *name, struct op *t)
 
 	if (t == NULL) {		/* undefine */
 		ktdelete(tp);
-		return was_set ? 0 : 1;
+		return (was_set ? 0 : 1);
 	}
 
 	tp->val.t = tcopy(t->left, tp->areap);
@@ -831,7 +876,7 @@ define(const char *name, struct op *t)
 	if (t->u.ksh_func)
 		tp->flag |= FKSH;
 
-	return 0;
+	return (0);
 }
 
 /*
@@ -869,9 +914,9 @@ struct tbl *
 findcom(const char *name, int flags)
 {
 	static struct tbl temp;
-	unsigned int h = hash(name);
+	uint32_t h = hash(name);
 	struct tbl *tp = NULL, *tbi;
-	int insert = Flag(FTRACKALL);	/* insert if not found */
+	unsigned char insert = Flag(FTRACKALL);	/* insert if not found */
 	char *fpath;			/* for function autoloading */
 	union mksh_cchack npath;
 
@@ -929,20 +974,15 @@ findcom(const char *name, int flags)
 		npath.ro = search(name, flags & FC_DEFPATH ? def_path : path,
 		    X_OK, &tp->u2.errno_);
 		if (npath.ro) {
-			/* XXX unsure about this, oksh does that
-			if (tp == &temp)
-				tp->val.s = npath.ro;
-			else */ {
-				tp->val.s = str_save(npath.ro, APERM);
-				if (npath.ro != name)
-					afree(npath.rw, ATEMP);
-			}
+			strdupx(tp->val.s, npath.ro, APERM);
+			if (npath.ro != name)
+				afree(npath.rw, ATEMP);
 			tp->flag |= ISSET|ALLOC;
 		} else if ((flags & FC_FUNC) &&
 		    (fpath = str_val(global("FPATH"))) != null &&
 		    (npath.ro = search(name, fpath, R_OK,
 		    &tp->u2.errno_)) != NULL) {
-			/* An undocumented feature of at&t ksh is that it
+			/* An undocumented feature of AT&T ksh is that it
 			 * searches FPATH if a command is not found, even
 			 * if the command hasn't been set up as an autoloaded
 			 * function (ie, no typeset -uf).
@@ -953,7 +993,7 @@ findcom(const char *name, int flags)
 			tp->u.fpath = npath.ro;
 		}
 	}
-	return tp;
+	return (tp);
 }
 
 /*
@@ -975,7 +1015,7 @@ flushcom(int all)	/* just relative or all */
 		}
 }
 
-/* Check if path is something we want to find.  Returns -1 for failure. */
+/* Check if path is something we want to find. Returns -1 for failure. */
 int
 search_access(const char *lpath, int mode,
     int *errnop)	/* set if candidate found, but not suitable */
@@ -984,19 +1024,19 @@ search_access(const char *lpath, int mode,
 	struct stat statb;
 
 	if (stat(lpath, &statb) < 0)
-		return -1;
+		return (-1);
 	ret = access(lpath, mode);
 	if (ret < 0)
 		err = errno; /* File exists, but we can't access it */
 	else if (mode == X_OK && (!S_ISREG(statb.st_mode) ||
 	    !(statb.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH)))) {
-	    /* This 'cause access() says root can execute everything */
+		/* This 'cause access() says root can execute everything */
 		ret = -1;
 		err = S_ISDIR(statb.st_mode) ? EISDIR : EACCES;
 	}
 	if (err && errnop && !*errnop)
 		*errnop = err;
-	return ret;
+	return (ret);
 }
 
 /*
@@ -1017,7 +1057,7 @@ search(const char *name, const char *lpath,
 	if (vstrchr(name, '/')) {
 		if (search_access(name, mode, errnop) == 0)
 			return (name);
-		return NULL;
+		return (NULL);
 	}
 
 	namelen = strlen(name) + 1;
@@ -1038,12 +1078,12 @@ search(const char *name, const char *lpath,
 		XcheckN(xs, xp, namelen);
 		memcpy(xp, name, namelen);
 		if (search_access(Xstring(xs, xp), mode, errnop) == 0)
-			return Xclose(xs, xp + namelen);
+			return (Xclose(xs, xp + namelen));
 		if (*sp++ == '\0')
 			sp = NULL;
 	}
 	Xfree(xs, xp);
-	return NULL;
+	return (NULL);
 }
 
 static int
@@ -1061,7 +1101,7 @@ call_builtin(struct tbl *tp, const char **wp)
 	shl_stdout_ok = 0;
 	builtin_flag = 0;
 	builtin_argv0 = NULL;
-	return rv;
+	return (rv);
 }
 
 /*
@@ -1120,8 +1160,7 @@ iosetup(struct ioword *iop, struct tbl *tp)
 		/* cp may have wrong name */
 		break;
 
-	case IODUP:
-	    {
+	case IODUP: {
 		const char *emsg;
 
 		do_open = 0;
@@ -1133,18 +1172,18 @@ iosetup(struct ioword *iop, struct tbl *tp)
 		    &emsg)) < 0) {
 			warningf(true, "%s: %s",
 			    snptreef(NULL, 32, "%R", &iotmp), emsg);
-			return -1;
+			return (-1);
 		}
 		if (u == iop->unit)
-			return 0;		/* "dup from" == "dup to" */
+			return (0);		/* "dup from" == "dup to" */
 		break;
-	    }
+	}
 	}
 
 	if (do_open) {
 		if (Flag(FRESTRICTED) && (flags & O_CREAT)) {
 			warningf(true, "%s: restricted", cp);
-			return -1;
+			return (-1);
 		}
 		u = open(cp, flags, 0666);
 	}
@@ -1157,7 +1196,7 @@ iosetup(struct ioword *iop, struct tbl *tp)
 			    (iotype == IOREAD || iotype == IOHERE) ?
 			    "open" : "create", cp, strerror(u));
 		}
-		return -1;
+		return (-1);
 	}
 	/* Do not save if it has already been redirected (i.e. "cat >x >y"). */
 	if (e->savefd[iop->unit] == 0) {
@@ -1166,7 +1205,7 @@ iosetup(struct ioword *iop, struct tbl *tp)
 			e->savefd[iop->unit] = -1;
 		else
 			/* c_exec() assumes e->savefd[fd] set for any
-			 * redirections.  Ask savefd() not to close iop->unit;
+			 * redirections. Ask savefd() not to close iop->unit;
 			 * this allows error messages to be seen if iop->unit
 			 * is 2; also means we can't lose the fd (eg, both
 			 * dup2 below and dup2 in restfd() failing).
@@ -1187,7 +1226,7 @@ iosetup(struct ioword *iop, struct tbl *tp)
 			    strerror(ev));
 			if (iotype != IODUP)
 				close(u);
-			return -1;
+			return (-1);
 		}
 		if (iotype != IODUP)
 			close(u);
@@ -1203,7 +1242,7 @@ iosetup(struct ioword *iop, struct tbl *tp)
 	}
 	if (u == 2) /* Clear any write errors */
 		shf_reopen(2, SHF_WR, shl_out);
-	return 0;
+	return (0);
 }
 
 /*
@@ -1222,7 +1261,7 @@ herein(const char *content, int sub)
 	/* ksh -c 'cat << EOF' can cause this... */
 	if (content == NULL) {
 		warningf(true, "here document missing");
-		return -2; /* special to iosetup(): don't print error */
+		return (-2); /* special to iosetup(): don't print error */
 	}
 
 	/* Create temp file to hold content (done before newenv so temp
@@ -1236,7 +1275,7 @@ herein(const char *content, int sub)
 		    h->name, strerror(fd));
 		if (shf)
 			shf_close(shf);
-		return -2 /* special to iosetup(): don't print error */;
+		return (-2 /* special to iosetup(): don't print error */);
 	}
 
 	osource = source;
@@ -1246,7 +1285,7 @@ herein(const char *content, int sub)
 		source = osource;
 		quitenv(shf);
 		close(fd);
-		return -2; /* special to iosetup(): don't print error */
+		return (-2); /* special to iosetup(): don't print error */
 	}
 	if (sub) {
 		/* Do substitutions on the content of heredoc */
@@ -1268,10 +1307,10 @@ herein(const char *content, int sub)
 		fd = errno;
 		warningf(true, "error writing %s: %s, %s", h->name,
 		    strerror(i), strerror(fd));
-		return -2; /* special to iosetup(): don't print error */
+		return (-2);	/* special to iosetup(): don't print error */
 	}
 
-	return fd;
+	return (fd);
 }
 
 /*
@@ -1299,121 +1338,128 @@ do_selectargs(const char **ap, bool print_menu)
 			pr_menu(ap);
 		shellf("%s", str_val(global("PS3")));
 		if (call_builtin(findcom("read", FC_BI), read_args))
-			return NULL;
+			return (NULL);
 		s = str_val(global("REPLY"));
 		if (*s) {
 			getn(s, &i);
-			return (i >= 1 && i <= argct) ? ap[i - 1] : null;
+			return ((i >= 1 && i <= argct) ? ap[i - 1] : null);
 		}
 		print_menu = 1;
 	}
 }
 
 struct select_menu_info {
-	const char *const *args;
-	int arg_width;
+	const char * const *args;
 	int num_width;
 };
 
-static char *select_fmt_entry(const void *, int, char *, int);
+static char *select_fmt_entry(char *, int, int, const void *);
 
 /* format a single select menu item */
 static char *
-select_fmt_entry(const void *arg, int i, char *buf, int buflen)
+select_fmt_entry(char *buf, int buflen, int i, const void *arg)
 {
 	const struct select_menu_info *smi =
 	    (const struct select_menu_info *)arg;
 
 	shf_snprintf(buf, buflen, "%*d) %s",
 	    smi->num_width, i + 1, smi->args[i]);
-	return buf;
+	return (buf);
 }
 
 /*
  *	print a select style menu
  */
 int
-pr_menu(const char *const *ap)
+pr_menu(const char * const *ap)
 {
 	struct select_menu_info smi;
-	const char *const *pp;
-	int nwidth, dwidth, i, n;
+	const char * const *pp;
+	int acols = 0, aocts = 0, i, n;
 
-	/* Width/column calculations were done once and saved, but this
-	 * means select can't be used recursively so we re-calculate each
-	 * time (could save in a structure that is returned, but its probably
-	 * not worth the bother).
+	/*
+	 * width/column calculations were done once and saved, but this
+	 * means select can't be used recursively so we re-calculate
+	 * each time (could save in a structure that is returned, but
+	 * it's probably not worth the bother)
 	 */
 
 	/*
 	 * get dimensions of the list
 	 */
-	for (n = 0, nwidth = 0, pp = ap; *pp; n++, pp++) {
+	for (n = 0, pp = ap; *pp; n++, pp++) {
+		i = strlen(*pp);
+		if (i > aocts)
+			aocts = i;
 		i = utf_mbswidth(*pp);
-		nwidth = (i > nwidth) ? i : nwidth;
+		if (i > acols)
+			acols = i;
 	}
+
 	/*
-	 * we will print an index of the form
-	 *	%d)
-	 * in front of each entry
-	 * get the max width of this
+	 * we will print an index of the form "%d) " in front of
+	 * each entry, so get the maximum width of this
 	 */
-	for (i = n, dwidth = 1; i >= 10; i /= 10)
-		dwidth++;
+	for (i = n, smi.num_width = 1; i >= 10; i /= 10)
+		smi.num_width++;
 
 	smi.args = ap;
-	smi.arg_width = nwidth;
-	smi.num_width = dwidth;
 	print_columns(shl_out, n, select_fmt_entry, (void *)&smi,
-	    dwidth + nwidth + 2, 1);
+	    smi.num_width + 2 + aocts, smi.num_width + 2 + acols,
+	    true);
 
-	return n;
+	return (n);
 }
 
 /* XXX: horrible kludge to fit within the framework */
-
-static char *plain_fmt_entry(const void *, int, char *, int);
+static char *plain_fmt_entry(char *, int, int, const void *);
 
 static char *
-plain_fmt_entry(const void *arg, int i, char *buf, int buflen)
+plain_fmt_entry(char *buf, int buflen, int i, const void *arg)
 {
-	shf_snprintf(buf, buflen, "%s", ((char *const *)arg)[i]);
-	return buf;
+	shf_snprintf(buf, buflen, "%s", ((const char * const *)arg)[i]);
+	return (buf);
 }
 
 int
-pr_list(char *const *ap)
+pr_list(char * const *ap)
 {
-	char *const *pp;
-	int nwidth, i, n;
+	int acols = 0, aocts = 0, i, n;
+	char * const *pp;
 
-	for (n = 0, nwidth = 0, pp = ap; *pp; n++, pp++) {
+	for (n = 0, pp = ap; *pp; n++, pp++) {
+		i = strlen(*pp);
+		if (i > aocts)
+			aocts = i;
 		i = utf_mbswidth(*pp);
-		nwidth = (i > nwidth) ? i : nwidth;
+		if (i > acols)
+			acols = i;
 	}
-	print_columns(shl_out, n, plain_fmt_entry, (const void *)ap,
-	    nwidth + 1, 0);
 
-	return n;
+	print_columns(shl_out, n, plain_fmt_entry, (const void *)ap,
+	    aocts, acols, false);
+
+	return (n);
 }
 
 /*
  *	[[ ... ]] evaluation routines
  */
 
-/* Test if the current token is a whatever.  Accepts the current token if
- * it is.  Returns 0 if it is not, non-zero if it is (in the case of
+/*
+ * Test if the current token is a whatever. Accepts the current token if
+ * it is. Returns 0 if it is not, non-zero if it is (in the case of
  * TM_UNOP and TM_BINOP, the returned value is a Test_op).
  */
-static int
+static Test_op
 dbteste_isa(Test_env *te, Test_meta meta)
 {
-	int ret = 0;
+	Test_op ret = TO_NONOP;
 	int uqword;
 	const char *p;
 
 	if (!*te->pos.wp)
-		return meta == TM_END;
+		return (meta == TM_END ? TO_NONNULL : TO_NONOP);
 
 	/* unquoted word? */
 	for (p = *te->pos.wp; *p == CHAR; p += 2)
@@ -1431,16 +1477,16 @@ dbteste_isa(Test_env *te, Test_meta meta)
 			ret = test_isop(meta, buf);
 		}
 	} else if (meta == TM_END)
-		ret = 0;
+		ret = TO_NONOP;
 	else
-		ret = uqword &&
-		    strcmp(*te->pos.wp, dbtest_tokens[(int) meta]) == 0;
+		ret = (uqword && !strcmp(*te->pos.wp,
+		    dbtest_tokens[(int)meta])) ? TO_NONNULL : TO_NONOP;
 
 	/* Accept the token? */
-	if (ret)
+	if (ret != TO_NONOP)
 		te->pos.wp++;
 
-	return ret;
+	return (ret);
 }
 
 static const char *
@@ -1449,19 +1495,19 @@ dbteste_getopnd(Test_env *te, Test_op op, bool do_eval)
 	const char *s = *te->pos.wp;
 
 	if (!s)
-		return NULL;
+		return (NULL);
 
 	te->pos.wp++;
 
 	if (!do_eval)
-		return null;
+		return (null);
 
 	if (op == TO_STEQL || op == TO_STNEQ)
 		s = evalstr(s, DOTILDE | DOPAT);
 	else
 		s = evalstr(s, DOTILDE);
 
-	return s;
+	return (s);
 }
 
 static void
