@@ -1,7 +1,27 @@
-/*	$OpenBSD: main.c,v 1.44 2008/07/05 07:25:18 djm Exp $	*/
+/*	$OpenBSD: main.c,v 1.46 2010/05/19 17:36:08 jasper Exp $	*/
 /*	$OpenBSD: tty.c,v 1.9 2006/03/14 22:08:01 deraadt Exp $	*/
 /*	$OpenBSD: io.c,v 1.22 2006/03/17 16:30:13 millert Exp $	*/
-/*	$OpenBSD: table.c,v 1.12 2005/12/11 20:31:21 otto Exp $	*/
+/*	$OpenBSD: table.c,v 1.13 2009/01/17 22:06:44 millert Exp $	*/
+
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
 
 #define	EXTERN
 #include "sh.h"
@@ -13,7 +33,7 @@
 #include <locale.h>
 #endif
 
-__RCSID("$MirOS: src/bin/mksh/main.c,v 1.101 2008/07/10 18:34:08 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/main.c,v 1.167 2010/07/04 17:45:15 tg Exp $");
 
 extern char **environ;
 
@@ -22,33 +42,51 @@ extern uid_t kshuid;
 extern gid_t kshgid, kshegid;
 #endif
 
+#ifndef MKSHRC_PATH
+#define MKSHRC_PATH	"~/.mkshrc"
+#endif
+
+#ifndef MKSH_DEFAULT_TMPDIR
+#define MKSH_DEFAULT_TMPDIR	"/tmp"
+#endif
+
 static void reclaim(void);
 static void remove_temps(struct temp *);
+void chvt_reinit(void);
+Source *mksh_init(int, const char *[]);
+#ifdef SIGWINCH
+static void x_sigwinch(int);
+#endif
 
 static const char initifs[] = "IFS= \t\n";
 
-static const char initsubs[] = "${PS2=> } ${PS3=#? } ${PS4=+ }";
+static const char initsubs[] =
+    "${PS2=> } ${PS3=#? } ${PS4=+ } ${SECONDS=0} ${TMOUT=0}";
 
 static const char *initcoms[] = {
-	"typeset", "-r", initvsn, NULL,
-	"typeset", "-x", "SHELL", "PATH", "HOME", NULL,
-	"typeset", "-i10", "OPTIND=1", "PGRP", "PPID", "USER_ID", NULL,
-	"eval", "typeset -i10 RANDOM SECONDS=\"${SECONDS-0}\" TMOUT=\"${TMOUT-0}\"", NULL,
-	"alias", "integer=typeset -i", "local=typeset", NULL,
+	T_typeset, "-r", initvsn, NULL,
+	T_typeset, "-x", "HOME", "PATH", "RANDOM", "SHELL", NULL,
+	T_typeset, "-i10", "COLUMNS", "LINES", "OPTIND", "PGRP", "PPID",
+	    "RANDOM", "SECONDS", "TMOUT", "USER_ID", NULL,
 	"alias",
+	"integer=typeset -i",
+	T_local_typeset,
 	"hash=alias -t",	/* not "alias -t --": hash -r needs to work */
 	"type=whence -v",
+#ifndef MKSH_UNEMPLOYED
 	"stop=kill -STOP",
 	"suspend=kill -STOP $$",
+#endif
 	"autoload=typeset -fu",
 	"functions=typeset -f",
 	"history=fc -l",
+	"nameref=typeset -n",
 	"nohup=nohup ",
-	"r=fc -e -",
+	r_fc_e_,
 	"source=PATH=$PATH:. command .",
 	"login=exec login",
 	NULL,
-	 /* this is what at&t ksh seems to track, with the addition of emacs */
+	 /* this is what AT&T ksh seems to track, with the addition of emacs */
 	"alias", "-tU",
 	"cat", "cc", "chmod", "cp", "date", "ed", "emacs", "grep", "ls",
 	"make", "mv", "pr", "rm", "sed", "sh", "vi", "who", NULL,
@@ -57,16 +95,25 @@ static const char *initcoms[] = {
 
 static int initio_done;
 
-int
-main(int argc, const char *argv[])
+struct env *e = &kshstate_v.env_;
+
+void
+chvt_reinit(void)
+{
+	kshpid = procpid = getpid();
+	ksheuid = geteuid();
+	kshpgrp = getpgrp();
+	kshppid = getppid();
+}
+
+Source *
+mksh_init(int argc, const char *argv[])
 {
 	int argi, i;
 	Source *s;
 	struct block *l;
-	int restricted, errexit;
+	unsigned char restricted, errexit, utf_flag;
 	const char **wp;
-	struct env env;
-	pid_t ppid;
 	struct tbl *vp;
 	struct stat s_stdin;
 #if !defined(_PATH_DEFPATH) && defined(_CS_PATH)
@@ -74,6 +121,9 @@ main(int argc, const char *argv[])
 	char *cp;
 #endif
 
+	/* do things like getpgrp() et al. */
+	chvt_reinit();
+
 	/* make sure argv[] is sane */
 	if (!*argv) {
 		static const char *empty_argv[] = {
@@ -88,10 +138,8 @@ main(int argc, const char *argv[])
 	ainit(&aperm);		/* initialise permanent Area */
 
 	/* set up base environment */
-	memset(&env, 0, sizeof(env));
-	env.type = E_NONE;
-	ainit(&env.area);
-	e = &env;
+	kshstate_v.env_.type = E_NONE;
+	ainit(&kshstate_v.env_.area);
 	newblock();		/* set up global l->vars and l->funs */
 
 	/* Do this first so output routines (eg, errorf, shellf) can work */
@@ -99,7 +147,7 @@ main(int argc, const char *argv[])
 
 	argi = parse_args(argv, OF_FIRSTTIME, NULL);
 	if (argi < 0)
-		exit(1);
+		return (NULL);
 
 	initvar();
 
@@ -112,7 +160,7 @@ main(int argc, const char *argv[])
 	/* set up variable and command dictionaries */
 	ktinit(&taliases, APERM, 0);
 	ktinit(&aliases, APERM, 0);
-#ifndef MKSH_SMALL
+#ifndef MKSH_NOPWNAM
 	ktinit(&homedirs, APERM, 0);
 #endif
 
@@ -120,7 +168,8 @@ main(int argc, const char *argv[])
 	initkeywords();
 
 	/* define built-in commands */
-	ktinit(&builtins, APERM, 64); /* must be 2^n (currently 40 builtins) */
+	ktinit(&builtins, APERM,
+	    /* must be 80% of 2^n (currently 44 builtins) */ 64);
 	for (i = 0; mkshbuiltins[i].name != NULL; i++)
 		builtin(mkshbuiltins[i].name, mkshbuiltins[i].func);
 
@@ -155,30 +204,45 @@ main(int argc, const char *argv[])
 
 	/* Turn on nohup by default for now - will change to off
 	 * by default once people are aware of its existence
-	 * (at&t ksh does not have a nohup option - it always sends
+	 * (AT&T ksh does not have a nohup option - it always sends
 	 * the hup).
 	 */
 	Flag(FNOHUP) = 1;
 
-	/* Turn on brace expansion by default.  AT&T kshs that have
+	/* Turn on brace expansion by default. AT&T kshs that have
 	 * alternation always have it on.
 	 */
 	Flag(FBRACEEXPAND) = 1;
 
 	/* Set edit mode to emacs by default, may be overridden
-	 * by the environment or the user.  Also, we want tab completion
+	 * by the environment or the user. Also, we want tab completion
 	 * on in vi by default. */
 	change_flag(FEMACS, OF_SPECIAL, 1);
-#ifndef MKSH_NOVI
+#if !MKSH_S_NOVI
 	Flag(FVITABCOMPLETE) = 1;
 #endif
 
+#ifdef MKSH_BINSHREDUCED
+	/* set FSH if we're called as -sh or /bin/sh or so */
+	{
+		const char *cc;
+
+		cc = kshname;
+		i = 0; argi = 0;
+		while (cc[i] != '\0')
+			/* the following line matches '-' and '/' ;-) */
+			if ((cc[i++] | 2) == '/')
+				argi = i;
+		if (((cc[argi] | 0x20) == 's') && ((cc[argi + 1] | 0x20) == 'h'))
+			change_flag(FSH, OF_FIRSTTIME, 1);
+	}
+#endif
+
 	/* import environment */
 	if (environ != NULL)
 		for (wp = (const char **)environ; *wp != NULL; wp++)
 			typeset(*wp, IMPORT | EXPORT, 0, 0, 0);
 
-	kshpid = procpid = getpid();
 	typeset(initifs, 0, 0, 0, 0);	/* for security */
 
 	/* assign default shell variable values */
@@ -207,29 +271,31 @@ main(int argc, const char *argv[])
 			/* setstr can't fail here */
 			setstr(pwd_v, current_wd, KSH_RETURN_ERROR);
 	}
-	ppid = getppid();
-	change_random(((unsigned long)kshname) ^
-	    ((unsigned long)time(NULL) * kshpid * ppid));
-#if HAVE_ARC4RANDOM
-	Flag(FARC4RANDOM) = 2;	/* use arc4random(3) until $RANDOM is written */
-#endif
 
 	for (wp = initcoms; *wp != NULL; wp++) {
 		shcomexec(wp);
 		while (*wp != NULL)
 			wp++;
 	}
+	setint(global("COLUMNS"), 0);
+	setint(global("LINES"), 0);
+	setint(global("OPTIND"), 1);
 
-	safe_prompt = (ksheuid = geteuid()) ? "$ " : "# ";
+	safe_prompt = ksheuid ? "$ " : "# ";
 	vp = global("PS1");
 	/* Set PS1 if unset or we are root and prompt doesn't contain a # */
 	if (!(vp->flag & ISSET) ||
 	    (!ksheuid && !strchr(str_val(vp), '#')))
 		/* setstr can't fail here */
 		setstr(vp, safe_prompt, KSH_RETURN_ERROR);
-	setint(global("PGRP"), (long)(kshpgrp = getpgrp()));
-	setint(global("PPID"), (long)ppid);
-	setint(global("USER_ID"), (long)ksheuid);
+	setint((vp = global("PGRP")), (mksh_uari_t)kshpgrp);
+	vp->flag |= INT_U;
+	setint((vp = global("PPID")), (mksh_uari_t)kshppid);
+	vp->flag |= INT_U;
+	setint((vp = global("RANDOM")), (mksh_uari_t)evilhash(kshname));
+	vp->flag |= INT_U;
+	setint((vp = global("USER_ID")), (mksh_uari_t)ksheuid);
+	vp->flag |= INT_U;
 
 	/* Set this before parsing arguments */
 #if HAVE_SETRESUGID
@@ -240,26 +306,44 @@ main(int argc, const char *argv[])
 #endif
 
 	/* this to note if monitor is set on command line (see below) */
+#ifndef MKSH_UNEMPLOYED
 	Flag(FMONITOR) = 127;
+#endif
+	/* this to note if utf-8 mode is set on command line (see below) */
+	UTFMODE = 2;
+
 	argi = parse_args(argv, OF_CMDLINE, NULL);
 	if (argi < 0)
-		exit(1);
+		return (NULL);
+
+	/* process this later only, default to off (hysterical raisins) */
+	utf_flag = UTFMODE;
+	UTFMODE = 0;
 
 	if (Flag(FCOMMAND)) {
 		s = pushs(SSTRING, ATEMP);
 		if (!(s->start = s->str = argv[argi++]))
 			errorf("-c requires an argument");
+#ifdef MKSH_MIDNIGHTBSD01ASH_COMPAT
+		/* compatibility to MidnightBSD 0.1 /bin/sh (kludge) */
+		if (Flag(FSH) && argv[argi] && !strcmp(argv[argi], "--"))
+			++argi;
+#endif
 		if (argv[argi])
 			kshname = argv[argi++];
 	} else if (argi < argc && !Flag(FSTDIN)) {
 		s = pushs(SFILE, ATEMP);
-		s->file = kshname = argv[argi++];
+		s->file = argv[argi++];
 		s->u.shf = shf_open(s->file, O_RDONLY, 0,
 		    SHF_MAPHI | SHF_CLEXEC);
 		if (s->u.shf == NULL) {
-			exstat = 127; /* POSIX */
-			errorf("%s: %s", s->file, strerror(errno));
+			shl_stdout_ok = 0;
+			warningf(true, "%s: %s", s->file, strerror(errno));
+			/* mandated by SUSv4 */
+			exstat = 127;
+			unwind(LERROR);
 		}
+		kshname = s->file;
 	} else {
 		Flag(FSTDIN) = 1;
 		s = pushs(SSTDIN, ATEMP);
@@ -275,22 +359,22 @@ main(int argc, const char *argv[])
 		}
 	}
 
-	/* This bizarreness is mandated by POSIX */
+	/* this bizarreness is mandated by POSIX */
 	if (fstat(0, &s_stdin) >= 0 && S_ISCHR(s_stdin.st_mode) &&
 	    Flag(FTALKING))
 		reset_nonblock(0);
 
 	/* initialise job control */
-	i = Flag(FMONITOR) != 127;
-	Flag(FMONITOR) = 0;
-	j_init(i);
-	/* Do this after j_init(), as tty_fd is not initialised 'til then */
+	j_init();
+	/* set: 0/1; unset: 2->0 */
+	UTFMODE = utf_flag & 1;
+	/* Do this after j_init(), as tty_fd is not initialised until then */
 	if (Flag(FTALKING)) {
+		if (utf_flag == 2) {
 #ifndef MKSH_ASSUME_UTF8
 #define isuc(x)	(((x) != NULL) && \
 		    (stristr((x), "UTF-8") || stristr((x), "utf8")))
 		/* Check if we're in a UTF-8 locale */
-		if (!Flag(FUTFHACK)) {
 			const char *ccp;
 
 #if HAVE_SETLOCALE_CTYPE
@@ -300,22 +384,30 @@ main(int argc, const char *argv[])
 				ccp = nl_langinfo(CODESET);
 #endif
 #else
-			ccp = getenv("LC_ALL");
-			if (!ccp || !*ccp) {
-				ccp = getenv("LC_CTYPE");
-				if (!ccp || !*ccp)
-					ccp = getenv("LANG");
-			}
+			/* these were imported from environ earlier */
+			ccp = str_val(global("LC_ALL"));
+			if (ccp == null)
+				ccp = str_val(global("LC_CTYPE"));
+			if (ccp == null)
+				ccp = str_val(global("LANG"));
 #endif
-			Flag(FUTFHACK) = isuc(ccp);
-		}
+			UTFMODE = isuc(ccp);
 #undef isuc
+#elif MKSH_ASSUME_UTF8
+			UTFMODE = 1;
 #else
-		Flag(FUTFHACK) = 1;
+			UTFMODE = 0;
 #endif
+		}
 		x_init();
 	}
 
+#ifdef SIGWINCH
+	sigtraps[SIGWINCH].flags |= TF_SHELL_USES;
+	setsig(&sigtraps[SIGWINCH], x_sigwinch,
+	    SS_RESTORE_ORIG|SS_FORCE|SS_SHTRAP);
+#endif
+
 	l = e->loc;
 	l->argv = &argv[argi - 1];
 	l->argc = argc - argi;
@@ -346,7 +438,7 @@ main(int argc, const char *argv[])
 		char *env_file;
 
 		/* include $ENV */
-		env_file = substitute(substitute("${ENV:-~/.mkshrc}", 0),
+		env_file = substitute(substitute("${ENV:-" MKSHRC_PATH "}", 0),
 		    DOTILDE);
 		if (*env_file != '\0')
 			include(env_file, 0, NULL, 1);
@@ -354,7 +446,7 @@ main(int argc, const char *argv[])
 
 	if (restricted) {
 		static const char *restr_com[] = {
-			"typeset", "-r", "PATH",
+			T_typeset, "-r", "PATH",
 			"ENV", "SHELL",
 			NULL
 		};
@@ -362,8 +454,7 @@ main(int argc, const char *argv[])
 		/* After typeset command... */
 		Flag(FRESTRICTED) = 1;
 	}
-	if (errexit)
-		Flag(FERREXIT) = 1;
+	Flag(FERREXIT) = errexit;
 
 	if (Flag(FTALKING)) {
 		hist_init(s);
@@ -371,8 +462,23 @@ main(int argc, const char *argv[])
 	} else
 		Flag(FTRACKALL) = 1;	/* set after ENV */
 
-	shell(s, true);	/* doesn't return */
-	return 0;
+	return (s);
+}
+
+int
+main(int argc, const char *argv[])
+{
+	Source *s;
+
+	kshstate_v.lcg_state_ = 5381;
+
+	if ((s = mksh_init(argc, argv))) {
+		/* put more entropy into the LCG */
+		change_random(s, sizeof(*s));
+		/* doesn’t return */
+		shell(s, true);
+	}
+	return (1);
 }
 
 int
@@ -386,7 +492,7 @@ include(const char *name, int argc, const char **argv, int intr_ok)
 
 	shf = shf_open(name, O_RDONLY, 0, SHF_MAPHI | SHF_CLEXEC);
 	if (shf == NULL)
-		return -1;
+		return (-1);
 
 	if (argv) {
 		old_argv = e->loc->argv;
@@ -406,14 +512,14 @@ include(const char *name, int argc, const char **argv, int intr_ok)
 		switch (i) {
 		case LRETURN:
 		case LERROR:
-			return exstat & 0xff; /* see below */
+			return (exstat & 0xff); /* see below */
 		case LINTR:
 			/* intr_ok is set if we are including .profile or $ENV.
 			 * If user ^Cs out, we don't want to kill the shell...
 			 */
 			if (intr_ok && (exstat - 128) != SIGTERM)
-				return 1;
-			/* FALLTHRU */
+				return (1);
+			/* FALLTHROUGH */
 		case LEXIT:
 		case LLEAVE:
 		case LSHELL:
@@ -430,24 +536,26 @@ include(const char *name, int argc, const char **argv, int intr_ok)
 	}
 	s = pushs(SFILE, ATEMP);
 	s->u.shf = shf;
-	s->file = str_save(name, ATEMP);
+	strdupx(s->file, name, ATEMP);
 	i = shell(s, false);
 	quitenv(s->u.shf);
 	if (old_argv) {
 		e->loc->argv = old_argv;
 		e->loc->argc = old_argc;
 	}
-	return i & 0xff;	/* & 0xff to ensure value not -1 */
+	return (i & 0xff);	/* & 0xff to ensure value not -1 */
 }
 
+/* spawn a command into a shell optionally keeping track of the line number */
 int
-command(const char *comm)
+command(const char *comm, int line)
 {
 	Source *s;
 
 	s = pushs(SSTRING, ATEMP);
 	s->start = s->str = comm;
-	return shell(s, false);
+	s->line = line;
+	return (shell(s, false));
 }
 
 /*
@@ -484,7 +592,7 @@ shell(Source * volatile s, volatile int toplevel)
 				    wastty)
 					s->type = SSTDIN;
 				/* Used by exit command to get back to
-				 * top level shell.  Kind of strange since
+				 * top level shell. Kind of strange since
 				 * interactive is set if we are reading from
 				 * a tty, but to have stopped jobs, one only
 				 * needs FMONITOR set (not FTALKING/SF_TTY)...
@@ -493,7 +601,7 @@ shell(Source * volatile s, volatile int toplevel)
 				s->start = s->str = null;
 				break;
 			}
-			/* FALLTHRU */
+			/* FALLTHROUGH */
 		case LEXIT:
 		case LLEAVE:
 		case LRETURN:
@@ -543,7 +651,7 @@ shell(Source * volatile s, volatile int toplevel)
 			}
 		}
 		if (t && (!Flag(FNOEXEC) || (s->flags & SF_TTY)))
-			exstat = execute(t, 0);
+			exstat = execute(t, 0, NULL);
 
 		if (t != NULL && t->type != TEOF && interactive && really_exit)
 			really_exit = 0;
@@ -552,14 +660,14 @@ shell(Source * volatile s, volatile int toplevel)
 	}
 	quitenv(NULL);
 	source = old_source;
-	return exstat;
+	return (exstat);
 }
 
 /* return to closest error handler or shell(), exit if none found */
 void
 unwind(int i)
 {
-	/* ordering for EXIT vs ERR is a bit odd (this is what at&t ksh does) */
+	/* ordering for EXIT vs ERR is a bit odd (this is what AT&T ksh does) */
 	if (i == LEXIT || (Flag(FERREXIT) && (i == LERROR || i == LINTR) &&
 	    sigtraps[SIGEXIT_].trap)) {
 		runtrap(&sigtraps[SIGEXIT_]);
@@ -580,7 +688,7 @@ unwind(int i)
 		case E_NONE:
 			if (i == LINTR)
 				e->flags |= EF_FAKE_SIGDIE;
-			/* FALLTHRU */
+			/* FALLTHROUGH */
 		default:
 			quitenv(NULL);
 		}
@@ -591,15 +699,23 @@ void
 newenv(int type)
 {
 	struct env *ep;
+	char *cp;
 
-	ep = (struct env *)alloc(sizeof (*ep), ATEMP);
-	ep->type = type;
-	ep->flags = 0;
+	/*
+	 * struct env includes ALLOC_ITEM for alignment constraints
+	 * so first get the actually used memory, then assign it
+	 */
+	cp = alloc(sizeof(struct env) - ALLOC_SIZE, ATEMP);
+	ep = (void *)(cp - ALLOC_SIZE);	/* undo what alloc() did */
+	/* initialise public members of struct env (not the ALLOC_ITEM) */
 	ainit(&ep->area);
+	ep->oenv = e;
 	ep->loc = e->loc;
 	ep->savefd = NULL;
-	ep->oenv = e;
 	ep->temps = NULL;
+	ep->type = type;
+	ep->flags = 0;
+	/* jump buffer is invalid because flags == 0 */
 	e = ep;
 }
 
@@ -607,6 +723,7 @@ void
 quitenv(struct shf *shf)
 {
 	struct env *ep = e;
+	char *cp;
 	int fd;
 
 	if (ep->oenv && ep->oenv->loc != ep->loc)
@@ -632,7 +749,7 @@ quitenv(struct shf *shf)
 			if (ep->flags & EF_FAKE_SIGDIE) {
 				int sig = exstat - 128;
 
-				/* ham up our death a bit (at&t ksh
+				/* ham up our death a bit (AT&T ksh
 				 * only seems to do this for SIGTERM)
 				 * Don't do it for SIGQUIT, since we'd
 				 * dump a core..
@@ -655,7 +772,10 @@ quitenv(struct shf *shf)
 	reclaim();
 
 	e = e->oenv;
-	afree(ep, ATEMP);
+
+	/* free the struct env - tricky due to the ALLOC_ITEM inside */
+	cp = (void *)ep;
+	afree(cp + ALLOC_SIZE, ATEMP);
 }
 
 /* Called after a fork to cleanup stuff left over from parents environment */
@@ -665,7 +785,10 @@ cleanup_parents_env(void)
 	struct env *ep;
 	int fd;
 
-	/* Don't clean up temporary files - parent will probably need them.
+	mkssert(e != NULL);
+
+	/*
+	 * Don't clean up temporary files - parent will probably need them.
 	 * Also, can't easily reclaim memory since variables, etc. could be
 	 * anywhere.
 	 */
@@ -710,13 +833,13 @@ remove_temps(struct temp *tp)
 			unlink(tp->name);
 }
 
-/* Initialise tty_fd.  Used for saving/reseting tty modes upon
+/* Initialise tty_fd. Used for saving/reseting tty modes upon
  * foreground job completion and for setting up tty process group.
  */
 void
-tty_init(int init_ttystate)
+tty_init(bool init_ttystate, bool need_tty)
 {
-	int do_close = 1;
+	bool do_close = true;
 	int tfd;
 
 	if (tty_fd >= 0) {
@@ -726,32 +849,40 @@ tty_init(int init_ttystate)
 	tty_devtty = 1;
 
 #ifdef _UWIN
+	/* XXX imake style */
 	if (isatty(3))
 		tfd = 3;
 	else
 #endif
 	if ((tfd = open("/dev/tty", O_RDWR, 0)) < 0) {
 		tty_devtty = 0;
-		warningf(false, "No controlling tty (open /dev/tty: %s)",
-		    strerror(errno));
+		if (need_tty)
+			warningf(false,
+			    "No controlling tty (open /dev/tty: %s)",
+			    strerror(errno));
 	}
 	if (tfd < 0) {
-		do_close = 0;
+		do_close = false;
 		if (isatty(0))
 			tfd = 0;
 		else if (isatty(2))
 			tfd = 2;
 		else {
-			warningf(false, "Can't find tty file descriptor");
+			if (need_tty)
+				warningf(false,
+				    "Can't find tty file descriptor");
 			return;
 		}
 	}
 	if ((tty_fd = fcntl(tfd, F_DUPFD, FDBASE)) < 0) {
-		warningf(false, "j_ttyinit: dup of tty fd failed: %s",
-		    strerror(errno));
+		if (need_tty)
+			warningf(false, "j_ttyinit: dup of tty fd failed: %s",
+			    strerror(errno));
 	} else if (fcntl(tty_fd, F_SETFD, FD_CLOEXEC) < 0) {
-		warningf(false, "j_ttyinit: can't set close-on-exec flag: %s",
-		    strerror(errno));
+		if (need_tty)
+			warningf(false,
+			    "j_ttyinit: can't set close-on-exec flag: %s",
+			    strerror(errno));
 		close(tty_fd);
 		tty_fd = -1;
 	} else if (init_ttystate)
@@ -777,7 +908,7 @@ errorf(const char *fmt, ...)
 
 	shl_stdout_ok = 0;	/* debugging: note that stdout not valid */
 	exstat = 1;
-	if (*fmt) {
+	if (*fmt != 1) {
 		error_prefix(true);
 		va_start(va, fmt);
 		shf_vfprintf(shl_out, fmt, va);
@@ -812,7 +943,7 @@ bi_errorf(const char *fmt, ...)
 
 	shl_stdout_ok = 0;	/* debugging: note that stdout not valid */
 	exstat = 1;
-	if (*fmt) {
+	if (*fmt != 1) {
 		error_prefix(true);
 		/* not set when main() calls parse_args() */
 		if (builtin_argv0)
@@ -834,9 +965,7 @@ bi_errorf(const char *fmt, ...)
 }
 
 /* Called when something that shouldn't happen does */
-static void internal_verrorf(const char *, va_list)
-    __attribute__((format (printf, 1, 0)));
-static void
+void
 internal_verrorf(const char *fmt, va_list ap)
 {
 	shf_fprintf(shl_out, "internal error: ");
@@ -914,8 +1043,8 @@ can_seek(int fd)
 {
 	struct stat statb;
 
-	return fstat(fd, &statb) == 0 && !S_ISREG(statb.st_mode) ?
-	    SHF_UNBUF : 0;
+	return (fstat(fd, &statb) == 0 && !S_ISREG(statb.st_mode) ?
+	    SHF_UNBUF : 0);
 }
 
 struct shf shf_iob[3];
@@ -939,6 +1068,7 @@ ksh_dup2(int ofd, int nfd, bool errok)
 		errorf("too many files open in shell");
 
 #ifdef __ultrix
+	/* XXX imake style */
 	if (rv >= 0)
 		fcntl(nfd, F_SETFD, 0);
 #endif
@@ -957,7 +1087,7 @@ savefd(int fd)
 
 	if (fd < FDBASE && (nfd = fcntl(fd, F_DUPFD, FDBASE)) < 0 &&
 	    errno == EBADF)
-		return -1;
+		return (-1);
 	if (nfd < 0 || nfd > SHRT_MAX)
 		errorf("too many files open in shell");
 	fcntl(nfd, F_SETFD, FD_CLOEXEC);
@@ -1024,7 +1154,7 @@ check_fd(const char *name, int mode, const char **emsgp)
 	fl &= O_ACCMODE;
 	/* X_OK is a kludge to disable this check for dups (x<&1):
 	 * historical shells never did this check (XXX don't know what
-	 * posix has to say).
+	 * POSIX has to say).
 	 */
 	if (!(mode & X_OK) && fl != O_RDWR && (
 	    ((mode & R_OK) && fl != O_RDONLY) ||
@@ -1091,10 +1221,10 @@ coproc_getfd(int mode, const char **emsgp)
 	int fd = (mode & R_OK) ? coproc.read : coproc.write;
 
 	if (fd >= 0)
-		return fd;
+		return (fd);
 	if (emsgp)
 		*emsgp = "no coprocess";
-	return -1;
+	return (-1);
 }
 
 /* called to close file descriptors related to the coprocess (if any)
@@ -1129,14 +1259,14 @@ maketemp(Area *ap, Temp_type type, struct temp **tlist)
 	char *pathname;
 	const char *dir;
 
-	dir = tmpdir ? tmpdir : "/tmp";
+	dir = tmpdir ? tmpdir : MKSH_DEFAULT_TMPDIR;
 #if HAVE_MKSTEMP
 	len = strlen(dir) + 6 + 10 + 1;
 #else
 	pathname = tempnam(dir, "mksh.");
 	len = ((pathname == NULL) ? 0 : strlen(pathname)) + 1;
 #endif
-	tp = (struct temp *)alloc(sizeof (struct temp) + len, ap);
+	tp = alloc(sizeof(struct temp) + len, ap);
 	tp->name = (char *)&tp[1];
 #if !HAVE_MKSTEMP
 	if (pathname == NULL)
@@ -1160,26 +1290,63 @@ maketemp(Area *ap, Temp_type type, struct temp **tlist)
 
 	tp->next = *tlist;
 	*tlist = tp;
-	return tp;
+	return (tp);
 }
 
+/*
+ * We use a similar collision resolution algorithm as Python 2.5.4
+ * but with a slightly tweaked implementation written from scratch.
+ */
+
 #define	INIT_TBLS	8	/* initial table size (power of 2) */
+#define PERTURB_SHIFT	5	/* see Python 2.5.4 Objects/dictobject.c */
 
-static void texpand(struct table *, int);
+static void texpand(struct table *, size_t);
 static int tnamecmp(const void *, const void *);
+static struct tbl *ktscan(struct table *, const char *, uint32_t,
+    struct tbl ***);
 
-unsigned int
-hash(const char *n)
+static void
+texpand(struct table *tp, size_t nsize)
 {
-	unsigned int h = 0;
+	size_t i, j, osize = tp->size, perturb;
+	struct tbl *tblp, **pp;
+	struct tbl **ntblp, **otblp = tp->tbls;
 
-	while (*n != '\0')
-		h = 2*h + *n++;
-	return h * 32821;	/* scatter bits */
+	ntblp = alloc(nsize * sizeof(struct tbl *), tp->areap);
+	for (i = 0; i < nsize; i++)
+		ntblp[i] = NULL;
+	tp->size = nsize;
+	tp->nfree = (nsize * 4) / 5;	/* table can get 80% full */
+	tp->tbls = ntblp;
+	if (otblp == NULL)
+		return;
+	nsize--;			/* from here on nsize := mask */
+	for (i = 0; i < osize; i++)
+		if ((tblp = otblp[i]) != NULL) {
+			if ((tblp->flag & DEFINED)) {
+				/* search for free hash table slot */
+				j = (perturb = tblp->ua.hval) & nsize;
+				goto find_first_empty_slot;
+ find_next_empty_slot:
+				j = (j << 2) + j + perturb + 1;
+				perturb >>= PERTURB_SHIFT;
+ find_first_empty_slot:
+				pp = &ntblp[j & nsize];
+				if (*pp != NULL)
+					goto find_next_empty_slot;
+				/* found an empty hash table slot */
+				*pp = tblp;
+				tp->nfree--;
+			} else if (!(tblp->flag & FINUSE)) {
+				afree(tblp, tp->areap);
+			}
+		}
+	afree(otblp, tp->areap);
 }
 
 void
-ktinit(struct table *tp, Area *ap, int tsize)
+ktinit(struct table *tp, Area *ap, size_t tsize)
 {
 	tp->areap = ap;
 	tp->tbls = NULL;
@@ -1188,66 +1355,41 @@ ktinit(struct table *tp, Area *ap, int tsize)
 		texpand(tp, tsize);
 }
 
-static void
-texpand(struct table *tp, int nsize)
-{
-	int i;
-	struct tbl *tblp, **p;
-	struct tbl **ntblp, **otblp = tp->tbls;
-	int osize = tp->size;
-
-	ntblp = (struct tbl **)alloc(sizeofN(struct tbl *, nsize), tp->areap);
-	for (i = 0; i < nsize; i++)
-		ntblp[i] = NULL;
-	tp->size = nsize;
-	tp->nfree = 8 * nsize / 10;	/* table can get 80% full */
-	tp->tbls = ntblp;
-	if (otblp == NULL)
-		return;
-	for (i = 0; i < osize; i++)
-		if ((tblp = otblp[i]) != NULL) {
-			if ((tblp->flag & DEFINED)) {
-				for (p = &ntblp[hash(tblp->name) &
-				    (tp->size - 1)]; *p != NULL; p--)
-					if (p == ntblp)	/* wrap */
-						p += tp->size;
-				*p = tblp;
-				tp->nfree--;
-			} else if (!(tblp->flag & FINUSE)) {
-				afree((void *)tblp, tp->areap);
-			}
-		}
-	afree((void *)otblp, tp->areap);
-}
-
-/* table */
-/* name to enter */
-/* hash(n) */
-struct tbl *
-ktsearch(struct table *tp, const char *n, unsigned int h)
+/* table, name (key) to search for, hash(name), rv pointer to tbl ptr */
+static struct tbl *
+ktscan(struct table *tp, const char *name, uint32_t h, struct tbl ***ppp)
 {
+	size_t j, perturb, mask;
 	struct tbl **pp, *p;
 
-	if (tp->size == 0)
-		return NULL;
-
-	/* search for name in hashed table */
-	for (pp = &tp->tbls[h & (tp->size - 1)]; (p = *pp) != NULL; pp--) {
-		if (*p->name == *n && strcmp(p->name, n) == 0 &&
-		    (p->flag & DEFINED))
-			return p;
-		if (pp == tp->tbls)	/* wrap */
-			pp += tp->size;
-	}
-
-	return NULL;
+	mask = tp->size - 1;
+	/* search for hash table slot matching name */
+	j = (perturb = h) & mask;
+	goto find_first_slot;
+ find_next_slot:
+	j = (j << 2) + j + perturb + 1;
+	perturb >>= PERTURB_SHIFT;
+ find_first_slot:
+	pp = &tp->tbls[j & mask];
+	if ((p = *pp) != NULL && (p->ua.hval != h || !(p->flag & DEFINED) ||
+	    strcmp(p->name, name)))
+		goto find_next_slot;
+	/* p == NULL if not found, correct found entry otherwise */
+	if (ppp)
+		*ppp = pp;
+	return (p);
 }
 
-/* table */
-/* name to enter */
-/* hash(n) */
+/* table, name (key) to search for, hash(n) */
 struct tbl *
-ktenter(struct table *tp, const char *n, unsigned int h)
+ktsearch(struct table *tp, const char *n, uint32_t h)
+{
+	return (tp->size ? ktscan(tp, n, h, NULL) : NULL);
+}
+
+/* table, name (key) to enter, hash(n) */
+struct tbl *
+ktenter(struct table *tp, const char *n, uint32_t h)
 {
 	struct tbl **pp, *p;
 	int len;
@@ -1255,25 +1397,22 @@ ktenter(struct table *tp, const char *n, unsigned int h)
 	if (tp->size == 0)
 		texpand(tp, INIT_TBLS);
  Search:
-	/* search for name in hashed table */
-	for (pp = &tp->tbls[h & (tp->size - 1)]; (p = *pp) != NULL; pp--) {
-		if (*p->name == *n && strcmp(p->name, n) == 0)
-			return p;	/* found */
-		if (pp == tp->tbls)	/* wrap */
-			pp += tp->size;
-	}
+	if ((p = ktscan(tp, n, h, &pp)))
+		return (p);
 
-	if (tp->nfree <= 0) {	/* too full */
+	if (tp->nfree <= 0) {
+		/* too full */
 		texpand(tp, 2 * tp->size);
 		goto Search;
 	}
+
 	/* create new tbl entry */
 	len = strlen(n) + 1;
-	p = (struct tbl *)alloc(offsetof(struct tbl, name[0])+len,
-	    tp->areap);
+	p = alloc(offsetof(struct tbl, name[0]) + len, tp->areap);
 	p->flag = 0;
 	p->type = 0;
 	p->areap = tp->areap;
+	p->ua.hval = h;
 	p->u2.field = 0;
 	p->u.array = NULL;
 	memcpy(p->name, n, len);
@@ -1281,7 +1420,7 @@ ktenter(struct table *tp, const char *n, unsigned int h)
 	/* enter in tp->tbls */
 	tp->nfree--;
 	*pp = p;
-	return p;
+	return (p);
 }
 
 void
@@ -1297,16 +1436,16 @@ ktnext(struct tstate *ts)
 	while (--ts->left >= 0) {
 		struct tbl *p = *ts->next++;
 		if (p != NULL && (p->flag & DEFINED))
-			return p;
+			return (p);
 	}
-	return NULL;
+	return (NULL);
 }
 
 static int
 tnamecmp(const void *p1, const void *p2)
 {
-	const struct tbl *a = *((struct tbl * const *)p1);
-	const struct tbl *b = *((struct tbl * const *)p2);
+	const struct tbl *a = *((const struct tbl * const *)p1);
+	const struct tbl *b = *((const struct tbl * const *)p2);
 
 	return (strcmp(a->name, b->name));
 }
@@ -1317,14 +1456,25 @@ ktsort(struct table *tp)
 	size_t i;
 	struct tbl **p, **sp, **dp;
 
-	p = (struct tbl **)alloc(sizeofN(struct tbl *, tp->size + 1), ATEMP);
+	p = alloc((tp->size + 1) * sizeof(struct tbl *), ATEMP);
 	sp = tp->tbls;		/* source */
 	dp = p;			/* dest */
-	for (i = 0; i < (size_t)tp->size; i++)
+	i = (size_t)tp->size;
+	while (i--)
 		if ((*dp = *sp++) != NULL && (((*dp)->flag & DEFINED) ||
 		    ((*dp)->flag & ARRAY)))
 			dp++;
-	qsort(p, (i = dp - p), sizeof (void *), tnamecmp);
+	qsort(p, (i = dp - p), sizeof(void *), tnamecmp);
 	p[i] = NULL;
-	return p;
+	return (p);
 }
+
+#ifdef SIGWINCH
+static void
+x_sigwinch(int sig MKSH_A_UNUSED)
+{
+	/* this runs inside interrupt context, with errno saved */
+
+	got_winch = 1;
+}
+#endif
