@@ -1,17 +1,35 @@
-/*	$OpenBSD: misc.c,v 1.34 2008/07/12 12:33:42 miod Exp $	*/
+/*	$OpenBSD: misc.c,v 1.37 2009/04/19 20:34:05 sthen Exp $	*/
 /*	$OpenBSD: path.c,v 1.12 2005/03/30 17:16:37 deraadt Exp $	*/
 
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
+
 #include "sh.h"
+#if !HAVE_GETRUSAGE
+#include <sys/times.h>
+#endif
 #if HAVE_GRP_H
 #include <grp.h>
 #endif
 
-__RCSID("$MirOS: src/bin/mksh/misc.c,v 1.85 2008/07/18 11:33:12 tg Exp $");
-
-#undef USE_CHVT
-#if defined(TIOCSCTTY) && !defined(MKSH_SMALL)
-#define USE_CHVT
-#endif
+__RCSID("$MirOS: src/bin/mksh/misc.c,v 1.141 2010/07/17 22:09:36 tg Exp $");
 
 unsigned char chtypes[UCHAR_MAX + 1];	/* type bits for unsigned char */
 
@@ -23,10 +41,9 @@ gid_t kshgid, kshegid;
 static int do_gmatch(const unsigned char *, const unsigned char *,
     const unsigned char *, const unsigned char *);
 static const unsigned char *cclass(const unsigned char *, int);
-#ifdef USE_CHVT
+#ifdef TIOCSCTTY
 static void chvt(const char *);
 #endif
-static char *do_phys_path(XString *, char *, const char *);
 
 /*
  * Fast character classes
@@ -60,37 +77,9 @@ initctypes(void)
 	setctypes("*@#!$-?", C_VAR1);
 	setctypes(" \t\n", C_IFSWS);
 	setctypes("=-+?", C_SUBOP1);
-	setctypes(" \n\t\"#$&'()*;<>?[]\\`|", C_QUOTE);
+	setctypes("\t\n \"#$&'()*;<=>?[\\]`|", C_QUOTE);
 }
 
-#if defined(MKSH_SMALL) || !HAVE_EXPSTMT
-char *
-str_nsave(const char *s, int n, Area *ap)
-{
-	char *rv = NULL;
-
-	if ((n >= 0) && (s != NULL))
-		strlcpy(rv = alloc(n + 1, ap), s, n + 1);
-	return (rv);
-}
-
-char *
-str_save(const char *s, Area *ap)
-{
-#ifdef MKSH_SMALL
-	return (s ? str_nsave(s, strlen(s), ap) : NULL);
-#else
-	char *rv = NULL;
-
-	if (s != NULL) {
-		size_t sz = strlen(s) + 1;
-		strlcpy(rv = alloc(sz, ap), s, sz);
-	}
-	return (rv);
-#endif
-}
-#endif
-
 /* called from XcheckN() to grow buffer */
 char *
 Xcheck_grow_(XString *xsp, const char *xp, unsigned int more)
@@ -100,56 +89,15 @@ Xcheck_grow_(XString *xsp, const char *xp, unsigned int more)
 	xsp->len += more > xsp->len ? more : xsp->len;
 	xsp->beg = aresize(xsp->beg, xsp->len + 8, xsp->areap);
 	xsp->end = xsp->beg + xsp->len;
-	return xsp->beg + (xp - old_beg);
+	return (xsp->beg + (xp - old_beg));
 }
 
+#define SHFLAGS_DEFNS
+#include "sh_flags.h"
+
 const struct shoption options[] = {
-	/* Special cases (see parse_args()): -A, -o, -s.
-	 * Options are sorted by their longnames - the order of these
-	 * entries MUST match the order of sh_flag F* enumerations in sh.h.
-	 */
-	{ "allexport",	'a',		OF_ANY },
-#if HAVE_ARC4RANDOM
-	{ "arc4random",	  0,		OF_ANY },
-#endif
-	{ "braceexpand",  0,		OF_ANY }, /* non-standard */
-	{ "bgnice",	  0,		OF_ANY },
-	{ NULL,		'c',	    OF_CMDLINE },
-	{ "emacs",	  0,		OF_ANY },
-	{ "errexit",	'e',		OF_ANY },
-	{ "gmacs",	  0,		OF_ANY },
-	{ "ignoreeof",	  0,		OF_ANY },
-	{ "interactive",'i',	    OF_CMDLINE },
-	{ "keyword",	'k',		OF_ANY },
-	{ "login",	'l',	    OF_CMDLINE },
-	{ "markdirs",	'X',		OF_ANY },
-	{ "monitor",	'm',		OF_ANY },
-	{ "noclobber",	'C',		OF_ANY },
-	{ "noexec",	'n',		OF_ANY },
-	{ "noglob",	'f',		OF_ANY },
-	{ "nohup",	  0,		OF_ANY },
-	{ "nolog",	  0,		OF_ANY }, /* no effect */
-	{ "notify",	'b',		OF_ANY },
-	{ "nounset",	'u',		OF_ANY },
-	{ "physical",	  0,		OF_ANY }, /* non-standard */
-	{ "posix",	  0,		OF_ANY }, /* non-standard */
-	{ "privileged",	'p',		OF_ANY },
-	{ "restricted",	'r',	    OF_CMDLINE },
-	{ "stdin",	's',	    OF_CMDLINE }, /* pseudo non-standard */
-	{ "trackall",	'h',		OF_ANY },
-	{ "utf8-hack",	'U',		OF_ANY }, /* non-standard */
-	{ "verbose",	'v',		OF_ANY },
-#ifndef MKSH_NOVI
-	{ "vi",		  0,		OF_ANY },
-	{ "viraw",	  0,		OF_ANY }, /* no effect */
-	{ "vi-tabcomplete",  0,		OF_ANY }, /* non-standard */
-	{ "vi-esccomplete",  0,		OF_ANY }, /* non-standard */
-#endif
-	{ "xtrace",	'x',		OF_ANY },
-	/* Anonymous flags: used internally by shell only
-	 * (not visible to user)
-	 */
-	{ NULL,		 0,	   OF_INTERNAL }, /* FTALKING_I */
+#define SHFLAGS_ITEMS
+#include "sh_flags.h"
 };
 
 /*
@@ -160,7 +108,11 @@ option(const char *n)
 {
 	size_t i;
 
-	for (i = 0; i < NELEM(options); i++)
+	if ((n[0] == '-' || n[0] == '+') && n[1] && !n[2]) {
+		for (i = 0; i < NELEM(options); i++)
+			if (options[i].c == n[1])
+				return (i);
+	} else for (i = 0; i < NELEM(options); i++)
 		if (options[i].name && strcmp(options[i].name, n) == 0)
 			return (i);
 
@@ -172,48 +124,56 @@ struct options_info {
 	int opts[NELEM(options)];
 };
 
-static char *options_fmt_entry(const void *arg, int, char *, int);
-static void printoptions(int);
+static char *options_fmt_entry(char *, int, int, const void *);
+static void printoptions(bool);
 
 /* format a single select menu item */
 static char *
-options_fmt_entry(const void *arg, int i, char *buf, int buflen)
+options_fmt_entry(char *buf, int buflen, int i, const void *arg)
 {
 	const struct options_info *oi = (const struct options_info *)arg;
 
 	shf_snprintf(buf, buflen, "%-*s %s",
 	    oi->opt_width, options[oi->opts[i]].name,
 	    Flag(oi->opts[i]) ? "on" : "off");
-	return buf;
+	return (buf);
 }
 
 static void
-printoptions(int verbose)
+printoptions(bool verbose)
 {
-	unsigned int i;
+	int i = 0;
 
 	if (verbose) {
+		int n = 0, len, octs = 0;
 		struct options_info oi;
-		int n, len;
 
 		/* verbose version */
 		shf_puts("Current option settings\n", shl_stdout);
 
-		for (i = n = oi.opt_width = 0; i < NELEM(options); i++)
+		oi.opt_width = 0;
+		while (i < (int)NELEM(options)) {
 			if (options[i].name) {
-				len = strlen(options[i].name);
 				oi.opts[n++] = i;
+				len = strlen(options[i].name);
+				if (len > octs)
+					octs = len;
+				len = utf_mbswidth(options[i].name);
 				if (len > oi.opt_width)
 					oi.opt_width = len;
 			}
+			++i;
+		}
 		print_columns(shl_stdout, n, options_fmt_entry, &oi,
-		    oi.opt_width + 5, 1);
+		    octs + 4, oi.opt_width + 4, true);
 	} else {
-		/* short version ala ksh93 */
+		/* short version á la AT&T ksh93 */
 		shf_puts("set", shl_stdout);
-		for (i = 0; i < NELEM(options); i++)
+		while (i < (int)NELEM(options)) {
 			if (Flag(i) && options[i].name)
 				shprintf(" -o %s", options[i].name);
+			++i;
+		}
 		shf_putc('\n', shl_stdout);
 	}
 }
@@ -228,32 +188,34 @@ getoptions(void)
 	for (i = 0; i < NELEM(options); i++)
 		if (options[i].c && Flag(i))
 			*cp++ = options[i].c;
-	return (str_nsave_(m, cp - m, ATEMP));
+	strndupx(cp, m, cp - m, ATEMP);
+	return (cp);
 }
 
 /* change a Flag(*) value; takes care of special actions */
 void
-change_flag(enum sh_flag f,
-    int what,		/* flag to change */
-    char newval)	/* what is changing the flag (command line vs set) */
+change_flag(enum sh_flag f, int what, unsigned int newval)
 {
-	char oldval;
+	unsigned char oldval;
 
 	oldval = Flag(f);
 	Flag(f) = newval ? 1 : 0;	/* needed for tristates */
+#ifndef MKSH_UNEMPLOYED
 	if (f == FMONITOR) {
 		if (what != OF_CMDLINE && newval != oldval)
 			j_change();
-	} else if ((
-#ifndef MKSH_NOVI
+	} else
+#endif
+	  if ((
+#if !MKSH_S_NOVI
 	    f == FVI ||
 #endif
 	    f == FEMACS || f == FGMACS) && newval) {
-#ifndef MKSH_NOVI
+#if !MKSH_S_NOVI
 		Flag(FVI) =
 #endif
 		    Flag(FEMACS) = Flag(FGMACS) = 0;
-		Flag(f) = newval;
+		Flag(f) = (unsigned char)newval;
 	} else if (f == FPRIVILEGED && oldval && !newval) {
 		/* Turning off -p? */
 #if HAVE_SETRESUGID
@@ -270,23 +232,24 @@ change_flag(enum sh_flag f,
 		setegid(kshegid = kshgid = getgid());
 		setgid(kshegid);
 #endif
-	} else if (f == FPOSIX && newval) {
-		Flag(FBRACEEXPAND) = 0;
+	} else if ((f == FPOSIX || f == FSH) && newval) {
+		Flag(FPOSIX) = Flag(FSH) = Flag(FBRACEEXPAND) = 0;
+		Flag(f) = (unsigned char)newval;
 	}
 	/* Changing interactive flag? */
 	if (f == FTALKING) {
 		if ((what == OF_CMDLINE || what == OF_SET) && procpid == kshpid)
-			Flag(FTALKING_I) = newval;
+			Flag(FTALKING_I) = (unsigned char)newval;
 	}
 }
 
-/* parse command line & set command arguments.  returns the index of
+/* Parse command line & set command arguments. Returns the index of
  * non-option arguments, -1 if there is an error.
  */
 int
 parse_args(const char **argv,
     int what,			/* OF_CMDLINE or OF_SET */
-    int *setargsp)
+    bool *setargsp)
 {
 	static char cmd_opts[NELEM(options) + 5]; /* o:T:\0 */
 	static char set_opts[NELEM(options) + 6]; /* A:o;s\0 */
@@ -296,14 +259,14 @@ parse_args(const char **argv,
 	size_t i;
 	int optc, sortargs = 0, arrayset = 0;
 
-	/* First call?  Build option strings... */
+	/* First call? Build option strings... */
 	if (cmd_opts[0] == '\0') {
 		char *p = cmd_opts, *q = set_opts;
 
 		/* see cmd_opts[] declaration */
 		*p++ = 'o';
 		*p++ = ':';
-#ifndef MKSH_SMALL
+#if !defined(MKSH_SMALL) || defined(TIOCSCTTY)
 		*p++ = 'T';
 		*p++ = ':';
 #endif
@@ -366,10 +329,16 @@ parse_args(const char **argv,
 				break;
 			}
 			i = option(go.optarg);
+			if ((enum sh_flag)i == FARC4RANDOM) {
+				warningf(true, "Do not use set ±o arc4random,"
+				    " it will be removed in the next version"
+				    " of mksh!");
+				return (0);
+			}
 			if ((i != (size_t)-1) && set == Flag(i))
 				/* Don't check the context if the flag
 				 * isn't changing - makes "set -o interactive"
-				 * work if you're already interactive.  Needed
+				 * work if you're already interactive. Needed
 				 * if the output of "set +o" is to be used.
 				 */
 				;
@@ -381,11 +350,11 @@ parse_args(const char **argv,
 			}
 			break;
 
-#ifndef MKSH_SMALL
+#if !defined(MKSH_SMALL) || defined(TIOCSCTTY)
 		case 'T':
 			if (what != OF_FIRSTTIME)
 				break;
-#ifndef USE_CHVT
+#ifndef TIOCSCTTY
 			errorf("no TIOCSCTTY ioctl");
 #else
 			change_flag(FTALKING, OF_CMDLINE, 1);
@@ -395,12 +364,12 @@ parse_args(const char **argv,
 #endif
 
 		case '?':
-			return -1;
+			return (-1);
 
 		default:
 			if (what == OF_FIRSTTIME)
 				break;
-			/* -s: sort positional params (at&t ksh stupidity) */
+			/* -s: sort positional params (AT&T ksh stupidity) */
 			if (what == OF_SET && optc == 's') {
 				sortargs = 1;
 				break;
@@ -431,21 +400,19 @@ parse_args(const char **argv,
 
 	if (arrayset && (!*array || *skip_varname(array, false))) {
 		bi_errorf("%s: is not an identifier", array);
-		return -1;
+		return (-1);
 	}
 	if (sortargs) {
 		for (i = go.optind; argv[i]; i++)
 			;
-		qsort(&argv[go.optind], i - go.optind, sizeof (void *),
+		qsort(&argv[go.optind], i - go.optind, sizeof(void *),
 		    xstrcmp);
 	}
-	if (arrayset) {
-		set_array(array, arrayset, argv + go.optind);
-		for (; argv[go.optind]; go.optind++)
-			;
-	}
+	if (arrayset)
+		go.optind += set_array(array, arrayset > 0 ? true : false,
+		    argv + go.optind);
 
-	return go.optind;
+	return (go.optind);
 }
 
 /* parse a decimal number: returns 0 if string isn't a number, 1 otherwise */
@@ -509,22 +476,22 @@ gmatchx(const char *s, const char *p, bool isfile)
 	const char *se, *pe;
 
 	if (s == NULL || p == NULL)
-		return 0;
+		return (0);
+
 	se = s + strlen(s);
 	pe = p + strlen(p);
 	/* isfile is false iff no syntax check has been done on
-	 * the pattern.  If check fails, just to a strcmp().
+	 * the pattern. If check fails, just to a strcmp().
 	 */
 	if (!isfile && !has_globbing(p, pe)) {
 		size_t len = pe - p + 1;
 		char tbuf[64];
-		char *t = len <= sizeof(tbuf) ? tbuf :
-		    (char *)alloc(len, ATEMP);
+		char *t = len <= sizeof(tbuf) ? tbuf : alloc(len, ATEMP);
 		debunk(t, p, len);
-		return !strcmp(t, s);
+		return (!strcmp(t, s));
 	}
-	return do_gmatch((const unsigned char *) s, (const unsigned char *) se,
-	    (const unsigned char *) p, (const unsigned char *) pe);
+	return (do_gmatch((const unsigned char *) s, (const unsigned char *) se,
+	    (const unsigned char *) p, (const unsigned char *) pe));
 }
 
 /* Returns if p is a syntacticly correct globbing pattern, false
@@ -571,7 +538,7 @@ has_globbing(const char *xp, const char *xpe)
 		} else if (c == ']') {
 			if (in_bracket) {
 				if (bnest)		/* [a*(b]) */
-					return 0;
+					return (0);
 				in_bracket = 0;
 			}
 		} else if ((c & 0x80) && vstrchr("*+?@! ", c & 0x7f)) {
@@ -582,18 +549,20 @@ has_globbing(const char *xp, const char *xpe)
 				nest++;
 		} else if (c == '|') {
 			if (in_bracket && !bnest)	/* *(a[foo|bar]) */
-				return 0;
+				return (0);
 		} else if (c == /*(*/ ')') {
 			if (in_bracket) {
 				if (!bnest--)		/* *(a[b)c] */
-					return 0;
+					return (0);
 			} else if (nest)
 				nest--;
 		}
-		/* else must be a MAGIC-MAGIC, or MAGIC-!, MAGIC--, MAGIC-]
-			 MAGIC-{, MAGIC-,, MAGIC-} */
+		/*
+		 * else must be a MAGIC-MAGIC, or MAGIC-!,
+		 * MAGIC--, MAGIC-], MAGIC-{, MAGIC-, MAGIC-}
+		 */
 	}
-	return saw_glob && !in_bracket && !nest;
+	return (saw_glob && !in_bracket && !nest);
 }
 
 /* Function must return either 0 or 1 (assumed by code for 0x80|'!') */
@@ -606,50 +575,54 @@ do_gmatch(const unsigned char *s, const unsigned char *se,
 	const unsigned char *srest;
 
 	if (s == NULL || p == NULL)
-		return 0;
+		return (0);
 	while (p < pe) {
 		pc = *p++;
 		sc = s < se ? *s : '\0';
 		s++;
 		if (!ISMAGIC(pc)) {
 			if (sc != pc)
-				return 0;
+				return (0);
 			continue;
 		}
 		switch (*p++) {
 		case '[':
 			if (sc == 0 || (p = cclass(p, sc)) == NULL)
-				return 0;
+				return (0);
 			break;
 
 		case '?':
 			if (sc == 0)
-				return 0;
+				return (0);
+			if (UTFMODE) {
+				--s;
+				s += utf_ptradj((const void *)s);
+			}
 			break;
 
 		case '*':
 			if (p == pe)
-				return 1;
+				return (1);
 			s--;
 			do {
 				if (do_gmatch(s, se, p, pe))
-					return 1;
+					return (1);
 			} while (s++ < se);
-			return 0;
+			return (0);
 
-		  /*
-		   * [*+?@!](pattern|pattern|..)
-		   * This is also needed for ${..%..}, etc.
-		   */
+		/**
+		 * [*+?@!](pattern|pattern|..)
+		 * This is also needed for ${..%..}, etc.
+		 */
 		case 0x80|'+': /* matches one or more times */
 		case 0x80|'*': /* matches zero or more times */
 			if (!(prest = pat_scan(p, pe, 0)))
-				return 0;
+				return (0);
 			s--;
 			/* take care of zero matches */
 			if (p[-1] == (0x80 | '*') &&
 			    do_gmatch(s, se, prest, pe))
-				return 1;
+				return (1);
 			for (psub = p; ; psub = pnext) {
 				pnext = pat_scan(psub, pe, 1);
 				for (srest = s; srest <= se; srest++) {
@@ -657,39 +630,39 @@ do_gmatch(const unsigned char *s, const unsigned char *se,
 					    (do_gmatch(srest, se, prest, pe) ||
 					    (s != srest && do_gmatch(srest,
 					    se, p - 2, pe))))
-						return 1;
+						return (1);
 				}
 				if (pnext == prest)
 					break;
 			}
-			return 0;
+			return (0);
 
 		case 0x80|'?': /* matches zero or once */
 		case 0x80|'@': /* matches one of the patterns */
 		case 0x80|' ': /* simile for @ */
 			if (!(prest = pat_scan(p, pe, 0)))
-				return 0;
+				return (0);
 			s--;
 			/* Take care of zero matches */
 			if (p[-1] == (0x80 | '?') &&
 			    do_gmatch(s, se, prest, pe))
-				return 1;
+				return (1);
 			for (psub = p; ; psub = pnext) {
 				pnext = pat_scan(psub, pe, 1);
 				srest = prest == pe ? se : s;
 				for (; srest <= se; srest++) {
 					if (do_gmatch(s, srest, psub, pnext - 2) &&
 					    do_gmatch(srest, se, prest, pe))
-						return 1;
+						return (1);
 				}
 				if (pnext == prest)
 					break;
 			}
-			return 0;
+			return (0);
 
 		case 0x80|'!': /* matches none of the patterns */
 			if (!(prest = pat_scan(p, pe, 0)))
-				return 0;
+				return (0);
 			s--;
 			for (srest = s; srest <= se; srest++) {
 				int matched = 0;
@@ -706,26 +679,26 @@ do_gmatch(const unsigned char *s, const unsigned char *se,
 				}
 				if (!matched &&
 				    do_gmatch(srest, se, prest, pe))
-					return 1;
+					return (1);
 			}
-			return 0;
+			return (0);
 
 		default:
 			if (sc != p[-1])
-				return 0;
+				return (0);
 			break;
 		}
 	}
-	return s == se;
+	return (s == se);
 }
 
 static const unsigned char *
 cclass(const unsigned char *p, int sub)
 {
-	int c, d, not, found = 0;
+	int c, d, notp, found = 0;
 	const unsigned char *orig_p = p;
 
-	if ((not = (ISMAGIC(*p) && *++p == NOT)))
+	if ((notp = (ISMAGIC(*p) && *++p == NOT)))
 		p++;
 	do {
 		c = *p++;
@@ -740,7 +713,7 @@ cclass(const unsigned char *p, int sub)
 		}
 		if (c == '\0')
 			/* No closing ] - act as if the opening [ was quoted */
-			return sub == '[' ? orig_p : NULL;
+			return (sub == '[' ? orig_p : NULL);
 		if (ISMAGIC(p[0]) && p[1] == '-' &&
 		    (!ISMAGIC(p[2]) || p[3] != ']')) {
 			p += 2; /* MAGIC- */
@@ -752,14 +725,14 @@ cclass(const unsigned char *p, int sub)
 			}
 			/* POSIX says this is an invalid expression */
 			if (c > d)
-				return NULL;
+				return (NULL);
 		} else
 			d = c;
 		if (c == sub || (c <= sub && sub <= d))
 			found = 1;
 	} while (!(ISMAGIC(p[0]) && p[1] == ']'));
 
-	return (found != not) ? p+2 : NULL;
+	return ((found != notp) ? p+2 : NULL);
 }
 
 /* Look for next ) or | (if match_sep) in *(foo|bar) pattern */
@@ -773,17 +746,17 @@ pat_scan(const unsigned char *p, const unsigned char *pe, int match_sep)
 			continue;
 		if ((*++p == /*(*/ ')' && nest-- == 0) ||
 		    (*p == '|' && match_sep && nest == 0))
-			return ++p;
+			return (p + 1);
 		if ((*p & 0x80) && vstrchr("*+?@! ", *p & 0x7f))
 			nest++;
 	}
-	return NULL;
+	return (NULL);
 }
 
 int
 xstrcmp(const void *p1, const void *p2)
 {
-	return (strcmp(*(char * const *)p1, *(char * const *)p2));
+	return (strcmp(*(const char * const *)p1, *(const char * const *)p2));
 }
 
 /* Initialise a Getopt structure */
@@ -815,11 +788,11 @@ ksh_getopt_reset(Getopt *go, int flags)
  *	  the option is missing).
  *	  Used for 'read -u2', 'print -u2' and fc -40.
  *	- '#' is like ':' in options, expect that the argument is optional
- *	  and must start with a digit.  If the argument doesn't start with a
+ *	  and must start with a digit. If the argument doesn't start with a
  *	  digit, it is assumed to be missing and normal option processing
  *	  continues (optarg is set to 0 if the option is missing).
  *	  Used for 'typeset -LZ4'.
- *	- accepts +c as well as -c IF the GF_PLUSOPT flag is present.  If an
+ *	- accepts +c as well as -c IF the GF_PLUSOPT flag is present. If an
  *	  option starting with + is accepted, the GI_PLUS flag will be set
  *	  in go->info.
  */
@@ -837,14 +810,14 @@ ksh_getopt(const char **argv, Getopt *go, const char *optionsp)
 			go->optind++;
 			go->p = 0;
 			go->info |= GI_MINUSMINUS;
-			return -1;
+			return (-1);
 		}
 		if (arg == NULL ||
 		    ((flag != '-' ) && /* neither a - nor a + (if + allowed) */
 		    (!(go->flags & GF_PLUSOPT) || flag != '+')) ||
 		    (c = arg[1]) == '\0') {
 			go->p = 0;
-			return -1;
+			return (-1);
 		}
 		go->optind++;
 		go->info &= ~(GI_MINUS|GI_PLUS);
@@ -863,7 +836,7 @@ ksh_getopt(const char **argv, Getopt *go, const char *optionsp)
 			if (go->flags & GF_ERROR)
 				bi_errorfz();
 		}
-		return '?';
+		return ('?');
 	}
 	/* : means argument must be present, may be part of option argument
 	 *   or the next argument
@@ -881,14 +854,14 @@ ksh_getopt(const char **argv, Getopt *go, const char *optionsp)
 			if (optionsp[0] == ':') {
 				go->buf[0] = c;
 				go->optarg = go->buf;
-				return ':';
+				return (':');
 			}
 			warningf(true, "%s%s-'%c' requires argument",
 			    (go->flags & GF_NONAME) ? "" : argv[0],
 			    (go->flags & GF_NONAME) ? "" : ": ", c);
 			if (go->flags & GF_ERROR)
 				bi_errorfz();
-			return '?';
+			return ('?');
 		}
 		go->p = 0;
 	} else if (*o == ',') {
@@ -897,7 +870,7 @@ ksh_getopt(const char **argv, Getopt *go, const char *optionsp)
 		go->p = 0;
 	} else if (*o == '#') {
 		/* argument is optional and may be attached or unattached
-		 * but must start with a digit.  optarg is set to 0 if the
+		 * but must start with a digit. optarg is set to 0 if the
 		 * argument is missing.
 		 */
 		if (argv[go->optind - 1][go->p]) {
@@ -914,7 +887,7 @@ ksh_getopt(const char **argv, Getopt *go, const char *optionsp)
 				go->optarg = NULL;
 		}
 	}
-	return c;
+	return (c);
 }
 
 /* print variable/alias value using necessary quotes
@@ -951,50 +924,71 @@ print_value_quoted(const char *s)
 		shf_putc('\'', shl_stdout);
 }
 
-/* Print things in columns and rows - func() is called to format the ith
- * element
+/*
+ * Print things in columns and rows - func() is called to format
+ * the i-th element
  */
 void
 print_columns(struct shf *shf, int n,
-    char *(*func) (const void *, int, char *, int),
-    const void *arg, int max_width, int prefcol)
+    char *(*func)(char *, int, int, const void *),
+    const void *arg, int max_oct, int max_col, bool prefcol)
 {
-	char *str = (char *)alloc(max_width + 1, ATEMP);
 	int i, r, c, rows, cols, nspace;
+	char *str;
 
-	/* max_width + 1 for the space.  Note that no space
-	 * is printed after the last column to avoid problems
-	 * with terminals that have auto-wrap.
-	 */
-	cols = x_cols / (max_width + 1);
-	if (!cols)
-		cols = 1;
-	rows = (n + cols - 1) / cols;
-	if (prefcol && n && cols > rows) {
-		int tmp = rows;
-
-		rows = cols;
-		cols = tmp;
-		if (rows > n)
-			rows = n;
+	if (n <= 0) {
+#ifndef MKSH_SMALL
+		internal_warningf("print_columns called with n=%d <= 0", n);
+#endif
+		return;
 	}
 
-	nspace = (x_cols - max_width * cols) / cols;
+	++max_oct;
+	str = alloc(max_oct, ATEMP);
+
+	/* ensure x_cols is valid first */
+	if (x_cols < MIN_COLS)
+		change_winsz();
+
+	/*
+	 * We use (max_col + 1) to consider the space separator.
+	 * Note that no space is printed after the last column
+	 * to avoid problems with terminals that have auto-wrap.
+	 */
+	cols = x_cols / (max_col + 1);
+
+	/* if we can only print one column anyway, skip the goo */
+	if (cols < 2) {
+		for (i = 0; i < n; ++i)
+			shf_fprintf(shf, "%s \n",
+			    (*func)(str, max_oct, i, arg));
+		goto out;
+	}
+
+	rows = (n + cols - 1) / cols;
+	if (prefcol && cols > rows) {
+		i = rows;
+		rows = cols > n ? n : cols;
+		cols = i;
+	}
+
+	max_col = -max_col;
+	nspace = (x_cols + max_col * cols) / cols;
 	if (nspace <= 0)
 		nspace = 1;
 	for (r = 0; r < rows; r++) {
 		for (c = 0; c < cols; c++) {
 			i = c * rows + r;
 			if (i < n) {
-				shf_fprintf(shf, "%-*s",
-				    max_width,
-				    (*func)(arg, i, str, max_width + 1));
+				shf_fprintf(shf, "%*s", max_col,
+				    (*func)(str, max_oct, i, arg));
 				if (c + 1 < cols)
 					shf_fprintf(shf, "%*s", nspace, null);
 			}
 		}
 		shf_putchar('\n', shf);
 	}
+ out:
 	afree(str, ATEMP);
 }
 
@@ -1044,7 +1038,7 @@ blocking_read(int fd, char *buf, int nbytes)
 		}
 		break;
 	}
-	return ret;
+	return (ret);
 }
 
 /* Reset the non-blocking flag on the specified file descriptor.
@@ -1057,13 +1051,13 @@ reset_nonblock(int fd)
 	int flags;
 
 	if ((flags = fcntl(fd, F_GETFL, 0)) < 0)
-		return -1;
+		return (-1);
 	if (!(flags & O_NONBLOCK))
-		return 0;
+		return (0);
 	flags &= ~O_NONBLOCK;
 	if (fcntl(fd, F_SETFL, flags) < 0)
-		return -1;
-	return 1;
+		return (-1);
+	return (1);
 }
 
 
@@ -1074,14 +1068,23 @@ ksh_get_wd(size_t *dlen)
 	char *ret, *b;
 	size_t len = 1;
 
+#ifdef NO_PATH_MAX
+	if ((b = get_current_dir_name())) {
+		len = strlen(b) + 1;
+		strndupx(ret, b, len - 1, ATEMP);
+		free(b);
+	} else
+		ret = NULL;
+#else
 	if ((ret = getcwd((b = alloc(PATH_MAX + 1, ATEMP)), PATH_MAX)))
 		ret = aresize(b, len = (strlen(b) + 1), ATEMP);
 	else
 		afree(b, ATEMP);
+#endif
 
 	if (dlen)
 		*dlen = len;
-	return ret;
+	return (ret);
 }
 
 /*
@@ -1105,19 +1108,18 @@ make_path(const char *cwd, const char *file,
     XString *xsp,
     int *phys_pathp)
 {
-	int	rval = 0;
-	int	use_cdpath = 1;
-	char	*plist;
-	int	len;
-	int	plen = 0;
-	char	*xp = Xstring(*xsp, xp);
+	int rval = 0;
+	bool use_cdpath = true;
+	char *plist;
+	int len, plen = 0;
+	char *xp = Xstring(*xsp, xp);
 
 	if (!file)
 		file = null;
 
 	if (file[0] == '/') {
 		*phys_pathp = 0;
-		use_cdpath = 0;
+		use_cdpath = false;
 	} else {
 		if (file[0] == '.') {
 			char c = file[1];
@@ -1125,22 +1127,22 @@ make_path(const char *cwd, const char *file,
 			if (c == '.')
 				c = file[2];
 			if (c == '/' || c == '\0')
-				use_cdpath = 0;
+				use_cdpath = false;
 		}
 
 		plist = *cdpathp;
 		if (!plist)
-			use_cdpath = 0;
+			use_cdpath = false;
 		else if (use_cdpath) {
 			char *pend;
 
 			for (pend = plist; *pend && *pend != ':'; pend++)
 				;
 			plen = pend - plist;
-			*cdpathp = *pend ? ++pend : NULL;
+			*cdpathp = *pend ? pend + 1 : NULL;
 		}
 
-		if ((use_cdpath == 0 || !plen || plist[0] != '/') &&
+		if ((!use_cdpath || !plen || plist[0] != '/') &&
 		    (cwd && *cwd)) {
 			len = strlen(cwd);
 			XcheckN(*xsp, xp, len);
@@ -1167,7 +1169,7 @@ make_path(const char *cwd, const char *file,
 	if (!use_cdpath)
 		*cdpathp = NULL;
 
-	return rval;
+	return (rval);
 }
 
 /*
@@ -1177,11 +1179,9 @@ make_path(const char *cwd, const char *file,
 void
 simplify_path(char *pathl)
 {
-	char	*cur;
-	char	*t;
-	int	isrooted;
-	char	*very_start = pathl;
-	char	*start;
+	char *cur, *t;
+	bool isrooted;
+	char *very_start = pathl, *start;
 
 	if (!*pathl)
 		return;
@@ -1190,13 +1190,13 @@ simplify_path(char *pathl)
 		very_start++;
 
 	/* Before			After
-	 *  /foo/			/foo
-	 *  /foo/../../bar		/bar
-	 *  /foo/./blah/..		/foo
-	 *  .				.
-	 *  ..				..
-	 *  ./foo			foo
-	 *  foo/../../../bar		../../bar
+	 * /foo/			/foo
+	 * /foo/../../bar		/bar
+	 * /foo/./blah/..		/foo
+	 * .				.
+	 * ..				..
+	 * ./foo			foo
+	 * foo/../../../bar		../../bar
 	 */
 
 	for (cur = t = start = very_start; ; ) {
@@ -1253,85 +1253,18 @@ set_current_wd(char *pathl)
 	} else
 		len = strlen(p) + 1;
 
-	if (len > current_wd_size)
-		current_wd = aresize(current_wd, current_wd_size = len, APERM);
+	if (len > current_wd_size) {
+		afree(current_wd, APERM);
+		current_wd = alloc(current_wd_size = len, APERM);
+	}
 	memcpy(current_wd, p, len);
 	if (p != pathl && p != null)
 		afree(p, ATEMP);
 }
 
-char *
-get_phys_path(const char *pathl)
-{
-	XString xs;
-	char *xp;
+#ifdef TIOCSCTTY
+extern void chvt_reinit(void);
 
-	Xinit(xs, xp, strlen(pathl) + 1, ATEMP);
-
-	xp = do_phys_path(&xs, xp, pathl);
-
-	if (!xp)
-		return NULL;
-
-	if (Xlength(xs, xp) == 0)
-		Xput(xs, xp, '/');
-	Xput(xs, xp, '\0');
-
-	return Xclose(xs, xp);
-}
-
-static char *
-do_phys_path(XString *xsp, char *xp, const char *pathl)
-{
-	const char *p, *q;
-	int len, llen;
-	int savepos;
-	char lbuf[PATH_MAX];
-
-	Xcheck(*xsp, xp);
-	for (p = pathl; p; p = q) {
-		while (*p == '/')
-			p++;
-		if (!*p)
-			break;
-		len = (q = cstrchr(p, '/')) ? q - p : (int)strlen(p);
-		if (len == 1 && p[0] == '.')
-			continue;
-		if (len == 2 && p[0] == '.' && p[1] == '.') {
-			while (xp > Xstring(*xsp, xp)) {
-				xp--;
-				if (*xp == '/')
-					break;
-			}
-			continue;
-		}
-
-		savepos = Xsavepos(*xsp, xp);
-		Xput(*xsp, xp, '/');
-		XcheckN(*xsp, xp, len + 1);
-		memcpy(xp, p, len);
-		xp += len;
-		*xp = '\0';
-
-		llen = readlink(Xstring(*xsp, xp), lbuf, sizeof(lbuf) - 1);
-		if (llen < 0) {
-			/* EINVAL means it wasn't a symlink... */
-			if (errno != EINVAL)
-				return NULL;
-			continue;
-		}
-		lbuf[llen] = '\0';
-
-		/* If absolute path, start from scratch.. */
-		xp = lbuf[0] == '/' ? Xstring(*xsp, xp) :
-		    Xrestpos(*xsp, xp, savepos);
-		if (!(xp = do_phys_path(xsp, xp, lbuf)))
-			return NULL;
-	}
-	return xp;
-}
-
-#ifdef USE_CHVT
 static void
 chvt(const char *fn)
 {
@@ -1339,15 +1272,18 @@ chvt(const char *fn)
 	struct stat sb;
 	int fd;
 
+	/* for entropy */
+	kshstate_f.h = evilhash(fn);
+
 	if (*fn == '-') {
-		memcpy(dv, "-/dev/null", sizeof ("-/dev/null"));
+		memcpy(dv, "-/dev/null", sizeof("-/dev/null"));
 		fn = dv + 1;
 	} else {
 		if (stat(fn, &sb)) {
 			memcpy(dv, "/dev/ttyC", 9);
-			strlcpy(dv + 9, fn, 20 - 9);
+			strlcpy(dv + 9, fn, sizeof(dv) - 9);
 			if (stat(dv, &sb)) {
-				strlcpy(dv + 8, fn, 20 - 8);
+				strlcpy(dv + 8, fn, sizeof(dv) - 8);
 				if (stat(dv, &sb))
 					errorf("chvt: can't find tty %s", fn);
 			}
@@ -1357,7 +1293,7 @@ chvt(const char *fn)
 			errorf("chvt: not a char device: %s", fn);
 		if ((sb.st_uid != 0) && chown(fn, 0, 0))
 			warningf(false, "chvt: cannot chown root %s", fn);
-		if (((sb.st_mode & 07777) != 0600) && chmod(fn, 0600))
+		if (((sb.st_mode & 07777) != 0600) && chmod(fn, (mode_t)0600))
 			warningf(false, "chvt: cannot chmod 0600 %s", fn);
 #if HAVE_REVOKE
 		if (revoke(fn))
@@ -1391,10 +1327,15 @@ chvt(const char *fn)
 	ksh_dup2(fd, 2, false);
 	if (fd > 2)
 		close(fd);
+	chvt_reinit();
 }
 #endif
 
 #ifdef DEBUG
+char longsizes_are_okay[sizeof(long) == sizeof(unsigned long) ? 1 : -1];
+char arisize_is_okay[sizeof(mksh_ari_t) == 4 ? 1 : -1];
+char uarisize_is_okay[sizeof(mksh_uari_t) == 4 ? 1 : -1];
+
 char *
 strchr(char *p, int ch)
 {
@@ -1448,11 +1389,191 @@ stristr(const char *b, const char *l)
 #endif
 #endif
 
-#if !HAVE_EXPSTMT
-bool
-ksh_isspace_(unsigned int ksh_isspace_c)
+#ifdef MKSH_SMALL
+char *
+strndup_(const char *src, size_t len, Area *ap)
 {
-	return ((ksh_isspace_c >= 0x09 && ksh_isspace_c <= 0x0D) ||
-	    (ksh_isspace_c == 0x20));
+	char *dst = NULL;
+
+	if (src != NULL) {
+		dst = alloc(len + 1, ap);
+		memcpy(dst, src, len);
+		dst[len] = '\0';
+	}
+	return (dst);
+}
+
+char *
+strdup_(const char *src, Area *ap)
+{
+	return (src == NULL ? NULL : strndup_(src, strlen(src), ap));
 }
 #endif
+
+#if !HAVE_GETRUSAGE
+#define INVTCK(r,t)	do {						\
+	r.tv_usec = ((t) % (1000000 / CLK_TCK)) * (1000000 / CLK_TCK);	\
+	r.tv_sec = (t) / CLK_TCK;					\
+} while (/* CONSTCOND */ 0)
+
+int
+getrusage(int what, struct rusage *ru)
+{
+	struct tms tms;
+	clock_t u, s;
+
+	if (/* ru == NULL || */ times(&tms) == (clock_t)-1)
+		return (-1);
+
+	switch (what) {
+	case RUSAGE_SELF:
+		u = tms.tms_utime;
+		s = tms.tms_stime;
+		break;
+	case RUSAGE_CHILDREN:
+		u = tms.tms_cutime;
+		s = tms.tms_cstime;
+		break;
+	default:
+		errno = EINVAL;
+		return (-1);
+	}
+	INVTCK(ru->ru_utime, u);
+	INVTCK(ru->ru_stime, s);
+	return (0);
+}
+#endif
+
+/*
+ * process the string available via fg (get a char)
+ * and fp (put back a char) for backslash escapes,
+ * assuming the first call to *fg gets the char di-
+ * rectly after the backslash; return the character
+ * (0..0xFF), Unicode (wc + 0x100), or -1 if no known
+ * escape sequence was found
+ */
+int
+unbksl(bool cstyle, int (*fg)(void), void (*fp)(int))
+{
+	int wc, i, c, fc;
+
+	fc = (*fg)();
+	switch (fc) {
+	case 'a':
+		/*
+		 * according to the comments in pdksh, \007 seems
+		 * to be more portable than \a (due to HP-UX cc,
+		 * Ultrix cc, old pcc, etc.) so we avoid the escape
+		 * sequence altogether in mksh and assume ASCII
+		 */
+		wc = 7;
+		break;
+	case 'b':
+		wc = '\b';
+		break;
+	case 'c':
+		if (!cstyle)
+			goto unknown_escape;
+		c = (*fg)();
+		wc = CTRL(c);
+		break;
+	case 'E':
+	case 'e':
+		wc = 033;
+		break;
+	case 'f':
+		wc = '\f';
+		break;
+	case 'n':
+		wc = '\n';
+		break;
+	case 'r':
+		wc = '\r';
+		break;
+	case 't':
+		wc = '\t';
+		break;
+	case 'v':
+		/* assume ASCII here as well */
+		wc = 11;
+		break;
+	case '1':
+	case '2':
+	case '3':
+	case '4':
+	case '5':
+	case '6':
+	case '7':
+		if (!cstyle)
+			goto unknown_escape;
+		/* FALLTHROUGH */
+	case '0':
+		if (cstyle)
+			(*fp)(fc);
+		/*
+		 * look for an octal number with up to three
+		 * digits, not counting the leading zero;
+		 * convert it to a raw octet
+		 */
+		wc = 0;
+		i = 3;
+		while (i--)
+			if ((c = (*fg)()) >= '0' && c <= '7')
+				wc = (wc << 3) + (c - '0');
+			else {
+				(*fp)(c);
+				break;
+			}
+		break;
+	case 'U':
+		i = 8;
+		if (0)
+		/* FALLTHROUGH */
+	case 'u':
+		i = 4;
+		if (0)
+		/* FALLTHROUGH */
+	case 'x':
+		i = cstyle ? -1 : 2;
+		/*
+		 * x:	look for a hexadecimal number with up to
+		 *	two (C style: arbitrary) digits; convert
+		 *	to raw octet (C style: Unicode if >0xFF)
+		 * u/U:	look for a hexadecimal number with up to
+		 *	four (U: eight) digits; convert to Unicode
+		 */
+		wc = 0;
+		while (i--) {
+			wc <<= 4;
+			if ((c = (*fg)()) >= '0' && c <= '9')
+				wc += c - '0';
+			else if (c >= 'A' && c <= 'F')
+				wc += c - 'A' + 10;
+			else if (c >= 'a' && c <= 'f')
+				wc += c - 'a' + 10;
+			else {
+				wc >>= 4;
+				(*fp)(c);
+				break;
+			}
+		}
+		if ((cstyle && wc > 0xFF) || fc != 'x')
+			/* Unicode marker */
+			wc += 0x100;
+		break;
+	case '\'':
+		if (!cstyle)
+			goto unknown_escape;
+		wc = '\'';
+		break;
+	case '\\':
+		wc = '\\';
+		break;
+	default:
+ unknown_escape:
+		(*fp)(fc);
+		return (-1);
+	}
+
+	return (wc);
+}
