@@ -1,8 +1,28 @@
-/*	$OpenBSD: syn.c,v 1.27 2006/04/10 14:38:59 jaredy Exp $	*/
+/*	$OpenBSD: syn.c,v 1.28 2008/07/23 16:34:38 jaredy Exp $	*/
+
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ */
 
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/syn.c,v 1.25 2008/07/18 11:33:13 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/syn.c,v 1.49 2010/07/17 22:09:39 tg Exp $");
 
 struct nesting_state {
 	int start_token;	/* token than began nesting (eg, FOR) */
@@ -21,22 +41,20 @@ static struct op *thenpart(void);
 static struct op *elsepart(void);
 static struct op *caselist(void);
 static struct op *casepart(int);
-static struct op *function_body(char *, int);
+static struct op *function_body(char *, bool);
 static char **wordlist(void);
 static struct op *block(int, struct op *, struct op *, char **);
 static struct op *newtp(int);
-static void syntaxerr(const char *)
-    __attribute__((noreturn));
+static void syntaxerr(const char *) MKSH_A_NORETURN;
 static void nesting_push(struct nesting_state *, int);
 static void nesting_pop(struct nesting_state *);
 static int assign_command(char *);
 static int inalias(struct source *);
-static int dbtestp_isa(Test_env *, Test_meta);
+static Test_op dbtestp_isa(Test_env *, Test_meta);
 static const char *dbtestp_getopnd(Test_env *, Test_op, bool);
 static int dbtestp_eval(Test_env *, Test_op, const char *,
     const char *, bool);
-static void dbtestp_error(Test_env *, int, const char *)
-    __attribute__((noreturn));
+static void dbtestp_error(Test_env *, int, const char *) MKSH_A_NORETURN;
 
 static struct op *outtree;		/* yyparse output */
 static struct nesting_state nesting;	/* \n changed to ; */
@@ -44,10 +62,10 @@ static struct nesting_state nesting;	/* \n changed to ; */
 static int reject;		/* token(cf) gets symbol again */
 static int symbol;		/* yylex value */
 
-#define	REJECT		(reject = 1)
-#define	ACCEPT		(reject = 0)
-#define	token(cf)	((reject) ? (ACCEPT, symbol) : (symbol = yylex(cf)))
-#define	tpeek(cf)	((reject) ? (symbol) : (REJECT, symbol = yylex(cf)))
+#define REJECT		(reject = 1)
+#define ACCEPT		(reject = 0)
+#define token(cf)	((reject) ? (ACCEPT, symbol) : (symbol = yylex(cf)))
+#define tpeek(cf)	((reject) ? (symbol) : (REJECT, symbol = yylex(cf)))
 #define musthave(c,cf)	do { if (token(cf) != (c)) syntaxerr(NULL); } while (0)
 
 static void
@@ -107,8 +125,7 @@ static struct op *
 c_list(int multi)
 {
 	struct op *t = NULL, *p, *tl = NULL;
-	int c;
-	int have_sep;
+	int c, have_sep;
 
 	while (1) {
 		p = andor();
@@ -124,7 +141,7 @@ c_list(int multi)
 			break;
 		else if (c == '&' || c == COPROC)
 			p = block(c == '&' ? TASYNC : TCOPROC,
-				  p, NOBLOCK, NOWORDS);
+			    p, NOBLOCK, NOWORDS);
 		else if (c != ';')
 			have_sep = 0;
 		if (!t)
@@ -137,7 +154,7 @@ c_list(int multi)
 			break;
 	}
 	REJECT;
-	return t;
+	return (t);
 }
 
 static struct ioword *
@@ -145,7 +162,7 @@ synio(int cf)
 {
 	struct ioword *iop;
 	static struct ioword *nextiop = NULL;
-	int ishere;
+	bool ishere;
 
 	if (nextiop != NULL) {
 		iop = nextiop;
@@ -163,19 +180,29 @@ synio(int cf)
 		iop->delim = yylval.cp;
 		if (*ident != 0) /* unquoted */
 			iop->flag |= IOEVAL;
-		if (herep >= &heres[HERES])
+		if (herep > &heres[HERES - 1])
 			yyerror("too many <<s\n");
 		*herep++ = iop;
 	} else
 		iop->name = yylval.cp;
 
 	if (iop->flag & IOBASH) {
-		nextiop = (struct ioword *)alloc(sizeof (*iop), ATEMP);
+		char *cp;
+
+		nextiop = alloc(sizeof(*iop), ATEMP);
+		nextiop->name = cp = alloc(5, ATEMP);
+
+		if (iop->unit > 9) {
+			*cp++ = CHAR;
+			*cp++ = '0' + (iop->unit / 10);
+		}
+		*cp++ = CHAR;
+		*cp++ = '0' + (iop->unit % 10);
+		*cp = EOS;
 
 		iop->flag &= ~IOBASH;
 		nextiop->unit = 2;
 		nextiop->flag = IODUP;
-		nextiop->name = shf_smprintf("%d", iop->unit);
 		nextiop->delim = NULL;
 		nextiop->heredoc = NULL;
 	}
@@ -204,8 +231,7 @@ get_command(int cf)
 	XPtrV args, vars;
 	struct nesting_state old_nesting;
 
-	iops = (struct ioword **)alloc(sizeofN(struct ioword *, NUFILE+1),
-	    ATEMP);
+	iops = alloc((NUFILE + 1) * sizeof(struct ioword *), ATEMP);
 	XPinit(args, 16);
 	XPinit(vars, 16);
 
@@ -213,10 +239,10 @@ get_command(int cf)
 	switch (c = token(cf|KEYWORD|ALIAS|VARASN)) {
 	default:
 		REJECT;
-		afree((void*) iops, ATEMP);
+		afree(iops, ATEMP);
 		XPfree(args);
 		XPfree(vars);
-		return NULL; /* empty line */
+		return (NULL); /* empty line */
 
 	case LWORD:
 	case REDIR:
@@ -239,7 +265,7 @@ get_command(int cf)
 			case LWORD:
 				ACCEPT;
 				/* the iopn == 0 and XPsize(vars) == 0 are
-				 * dubious but at&t ksh acts this way
+				 * dubious but AT&T ksh acts this way
 				 */
 				if (iopn == 0 && XPsize(vars) == 0 &&
 				    XPsize(args) == 0 &&
@@ -253,7 +279,7 @@ get_command(int cf)
 				break;
 
 			case '(':
-				/* Check for "> foo (echo hi)" which at&t ksh
+				/* Check for "> foo (echo hi)" which AT&T ksh
 				 * allows (not POSIX, but not disallowed)
 				 */
 				afree(t, ATEMP);
@@ -261,9 +287,11 @@ get_command(int cf)
 					ACCEPT;
 					goto Subshell;
 				}
+#ifndef MKSH_SMALL
 				if ((XPsize(args) == 0 || Flag(FKEYWORD)) &&
 				    XPsize(vars) == 1 && is_wdvarassign(yylval.cp))
 					goto is_wdarrassign;
+#endif
 				/* Must be a function */
 				if (iopn != 0 || XPsize(args) != 1 ||
 				    XPsize(vars) != 0)
@@ -273,8 +301,9 @@ get_command(int cf)
 				musthave(')', 0);
 				t = function_body(XPptrv(args)[0], false);
 				goto Leave;
+#ifndef MKSH_SMALL
  is_wdarrassign:
-			  {
+			{
 				static const char set_cmd0[] = {
 					CHAR, 'e', CHAR, 'v',
 					CHAR, 'a', CHAR, 'l', EOS
@@ -307,7 +336,8 @@ get_command(int cf)
 				musthave(LWORD,LETARRAY);
 				XPput(args, yylval.cp);
 				break;
-			  }
+			}
+#endif
 
 			default:
 				goto Leave;
@@ -316,8 +346,8 @@ get_command(int cf)
  Leave:
 		break;
 
- Subshell:
 	case '(':
+ Subshell:
 		t = nested(TPAREN, '(', ')');
 		break;
 
@@ -325,21 +355,30 @@ get_command(int cf)
 		t = nested(TBRACE, '{', '}');
 		break;
 
-	case MDPAREN:
-	  {
+	case MDPAREN: {
+		int lno;
 		static const char let_cmd[] = {
 			CHAR, 'l', CHAR, 'e',
 			CHAR, 't', EOS
 		};
+
 		/* Leave KEYWORD in syniocf (allow if (( 1 )) then ...) */
-		t = newtp(TCOM);
-		t->lineno = source->line;
+		lno = source->line;
 		ACCEPT;
+		switch (token(LETEXPR)) {
+		case LWORD:
+			break;
+		case '(':	/* ) */
+			goto Subshell;
+		default:
+			syntaxerr(NULL);
+		}
+		t = newtp(TCOM);
+		t->lineno = lno;
 		XPput(args, wdcopy(let_cmd, ATEMP));
-		musthave(LWORD,LETEXPR);
 		XPput(args, yylval.cp);
 		break;
-	  }
+	}
 
 	case DBRACKET: /* [[ .. ]] */
 		/* Leave KEYWORD in syniocf (allow if [[ -n 1 ]] then ...) */
@@ -366,7 +405,7 @@ get_command(int cf)
 		if (!is_wdvarname(yylval.cp, true))
 			yyerror("%s: bad identifier\n",
 			    c == FOR ? "for" : "select");
-		t->str = str_save_(ident, ATEMP);
+		strdupx(t->str, ident, ATEMP);
 		nesting_push(&old_nesting, c);
 		t->vars = wordlist();
 		t->left = dogroup();
@@ -413,7 +452,7 @@ get_command(int cf)
 		t = pipeline(0);
 		if (t) {
 			t->str = alloc(2, ATEMP);
-			t->str[0] = 0;	/* TF_* flags */
+			t->str[0] = '\0';	/* TF_* flags */
 			t->str[1] = '\0';
 		}
 		t = block(TTIME, t, NOBLOCK, NOWORDS);
@@ -432,12 +471,11 @@ get_command(int cf)
 	}
 
 	if (iopn == 0) {
-		afree((void*)iops, ATEMP);
+		afree(iops, ATEMP);
 		t->ioact = NULL;
 	} else {
 		iops[iopn++] = NULL;
-		iops = (struct ioword **) aresize((void*) iops,
-		    sizeofN(struct ioword *, iopn), ATEMP);
+		iops = aresize(iops, iopn * sizeof(struct ioword *), ATEMP);
 		t->ioact = iops;
 	}
 
@@ -451,7 +489,7 @@ get_command(int cf)
 		XPfree(vars);
 	}
 
-	return t;
+	return (t);
 }
 
 static struct op *
@@ -474,7 +512,7 @@ dogroup(void)
 		syntaxerr(NULL);
 	list = c_list(true);
 	musthave(c, KEYWORD|ALIAS);
-	return list;
+	return (list);
 }
 
 static struct op *
@@ -511,7 +549,7 @@ elsepart(void)
 	default:
 		REJECT;
 	}
-	return NULL;
+	return (NULL);
 }
 
 static struct op *
@@ -544,25 +582,24 @@ static struct op *
 casepart(int endtok)
 {
 	struct op *t;
-	int c;
 	XPtrV ptns;
 
 	XPinit(ptns, 16);
 	t = newtp(TPAT);
-	c = token(CONTIN|KEYWORD); /* no ALIAS here */
-	if (c != '(')
+	/* no ALIAS here */
+	if (token(CONTIN | KEYWORD) != '(')
 		REJECT;
 	do {
 		musthave(LWORD, 0);
 		XPput(ptns, yylval.cp);
-	} while ((c = token(0)) == '|');
+	} while (token(0) == '|');
 	REJECT;
 	XPput(ptns, NULL);
 	t->vars = (char **) XPclose(ptns);
 	musthave(')', 0);
 
 	t->left = c_list(true);
-	/* Note: Posix requires the ;; */
+	/* Note: POSIX requires the ;; */
 	if ((tpeek(CONTIN|KEYWORD|ALIAS)) != endtok)
 		musthave(BREAK, CONTIN|KEYWORD|ALIAS);
 	return (t);
@@ -570,38 +607,49 @@ casepart(int endtok)
 
 static struct op *
 function_body(char *name,
-    int ksh_func)		/* function foo { ... } vs foo() { .. } */
+    bool ksh_func)		/* function foo { ... } vs foo() { .. } */
 {
 	char *sname, *p;
 	struct op *t;
-	int old_func_parse;
+	bool old_func_parse;
 
 	sname = wdstrip(name, false, false);
-	/* Check for valid characters in name.  posix and ksh93 say only
+	/* Check for valid characters in name. POSIX and AT&T ksh93 say only
 	 * allow [a-zA-Z_0-9] but this allows more as old pdkshs have
 	 * allowed more (the following were never allowed:
-	 *	nul space nl tab $ ' " \ ` ( ) & | ; = < >
-	 *  C_QUOTE covers all but = and adds # [ ] ? *)
+	 *	NUL TAB NL SP " $ & ' ( ) ; < = > \ ` |
+	 * C_QUOTE covers all but adds # * ? [ ]
 	 */
 	for (p = sname; *p; p++)
-		if (ctype(*p, C_QUOTE) || *p == '=')
+		if (ctype(*p, C_QUOTE))
 			yyerror("%s: invalid function name\n", sname);
 
+	/* Note that POSIX allows only compound statements after foo(), sh and
+	 * AT&T ksh allow any command, go with the later since it shouldn't
+	 * break anything. However, for function foo, AT&T ksh only accepts
+	 * an open-brace.
+	 */
+	if (ksh_func) {
+		if (tpeek(CONTIN|KEYWORD|ALIAS) == '(' /* ) */) {
+			struct tbl *tp;
+
+			/* function foo () { */
+			ACCEPT;
+			musthave(')', 0);
+			/* degrade to POSIX function */
+			ksh_func = false;
+			if ((tp = ktsearch(&aliases, sname, hash(sname))))
+				ktdelete(tp);
+		}
+		musthave('{', CONTIN|KEYWORD|ALIAS); /* } */
+		REJECT;
+	}
+
 	t = newtp(TFUNCT);
 	t->str = sname;
 	t->u.ksh_func = ksh_func;
 	t->lineno = source->line;
 
-	/* Note that POSIX allows only compound statements after foo(), sh and
-	 * at&t ksh allow any command, go with the later since it shouldn't
-	 * break anything.  However, for function foo, at&t ksh only accepts
-	 * an open-brace.
-	 */
-	if (ksh_func) {
-		musthave('{', CONTIN|KEYWORD|ALIAS); /* } */
-		REJECT;
-	}
-
 	old_func_parse = e->flags & EF_FUNC_PARSE;
 	e->flags |= EF_FUNC_PARSE;
 	if ((t->left = get_command(CONTIN)) == NULL) {
@@ -613,21 +661,20 @@ function_body(char *name,
 		 * be used as input), we pretend there is a colon here.
 		 */
 		t->left = newtp(TCOM);
-		t->left->args = (const char **)alloc(sizeof (char *) * 2,
-		    ATEMP);
-		t->left->args[0] = tv = alloc(sizeof (char) * 3, ATEMP);
+		t->left->args = alloc(2 * sizeof(char *), ATEMP);
+		t->left->args[0] = tv = alloc(3, ATEMP);
 		tv[0] = CHAR;
 		tv[1] = ':';
 		tv[2] = EOS;
 		t->left->args[1] = NULL;
-		t->left->vars = (char **)alloc(sizeof (char *), ATEMP);
+		t->left->vars = alloc(sizeof(char *), ATEMP);
 		t->left->vars[0] = NULL;
 		t->left->lineno = 1;
 	}
 	if (!old_func_parse)
 		e->flags &= ~EF_FUNC_PARSE;
 
-	return t;
+	return (t);
 }
 
 static char **
@@ -637,11 +684,11 @@ wordlist(void)
 	XPtrV args;
 
 	XPinit(args, 16);
-	/* Posix does not do alias expansion here... */
+	/* POSIX does not do alias expansion here... */
 	if ((c = token(CONTIN|KEYWORD|ALIAS)) != IN) {
-		if (c != ';') /* non-POSIX, but at&t ksh accepts a ; here */
+		if (c != ';') /* non-POSIX, but AT&T ksh accepts a ; here */
 			REJECT;
-		return NULL;
+		return (NULL);
 	}
 	while ((c = token(0)) == LWORD)
 		XPput(args, yylval.cp);
@@ -649,10 +696,10 @@ wordlist(void)
 		syntaxerr(NULL);
 	if (XPsize(args) == 0) {
 		XPfree(args);
-		return NULL;
+		return (NULL);
 	} else {
 		XPput(args, NULL);
-		return (char **) XPclose(args);
+		return ((char **)XPclose(args));
 	}
 }
 
@@ -672,10 +719,10 @@ block(int type, struct op *t1, struct op *t2, char **wp)
 	return (t);
 }
 
-const	struct tokeninfo {
+const struct tokeninfo {
 	const char *name;
-	short	val;
-	short	reserved;
+	short val;
+	short reserved;
 } tokentab[] = {
 	/* Reserved words */
 	{ "if",		IF,	true },
@@ -715,7 +762,8 @@ initkeywords(void)
 	struct tokeninfo const *tt;
 	struct tbl *p;
 
-	ktinit(&keywords, APERM, 32); /* must be 2^n (currently 20 keywords) */
+	ktinit(&keywords, APERM,
+	    /* must be 80% of 2^n (currently 20 keywords) */ 32);
 	for (tt = tokentab; tt->name; tt++) {
 		if (tt->reserved) {
 			p = ktenter(&keywords, tt->name, hash(tt->name));
@@ -797,7 +845,7 @@ newtp(int type)
 {
 	struct op *t;
 
-	t = (struct op *)alloc(sizeof (*t), ATEMP);
+	t = alloc(sizeof(struct op), ATEMP);
 	t->type = type;
 	t->u.evalflags = 0;
 	t->args = NULL;
@@ -816,13 +864,13 @@ compile(Source *s)
 	herep = heres;
 	source = s;
 	yyparse();
-	return outtree;
+	return (outtree);
 }
 
-/* This kludge exists to take care of sh/at&t ksh oddity in which
+/* This kludge exists to take care of sh/AT&T ksh oddity in which
  * the arguments of alias/export/readonly/typeset have no field
  * splitting, file globbing, or (normal) tilde expansion done.
- * at&t ksh seems to do something similar to this since
+ * AT&T ksh seems to do something similar to this since
  *	$ touch a=a; typeset a=[ab]; echo "$a"
  *	a=[ab]
  *	$ x=typeset; $x a=[ab]; echo "$a"
@@ -833,11 +881,11 @@ static int
 assign_command(char *s)
 {
 	if (!*s)
-		return 0;
-	return (strcmp(s, "alias") == 0) ||
+		return (0);
+	return ((strcmp(s, "alias") == 0) ||
 	    (strcmp(s, "export") == 0) ||
 	    (strcmp(s, "readonly") == 0) ||
-	    (strcmp(s, "typeset") == 0);
+	    (strcmp(s, T_typeset) == 0));
 }
 
 /* Check if we are in the middle of reading an alias */
@@ -846,8 +894,8 @@ inalias(struct source *s)
 {
 	for (; s && s->type == SALIAS; s = s->next)
 		if (!(s->flags & SF_ALIASEND))
-			return 1;
-	return 0;
+			return (1);
+	return (0);
 }
 
 
@@ -869,71 +917,75 @@ const char db_close[] = { CHAR, ']', CHAR, ']', EOS };
 const char db_lthan[] = { CHAR, '<', EOS };
 const char db_gthan[] = { CHAR, '>', EOS };
 
-/* Test if the current token is a whatever.  Accepts the current token if
- * it is.  Returns 0 if it is not, non-zero if it is (in the case of
+/*
+ * Test if the current token is a whatever. Accepts the current token if
+ * it is. Returns 0 if it is not, non-zero if it is (in the case of
  * TM_UNOP and TM_BINOP, the returned value is a Test_op).
  */
-static int
+static Test_op
 dbtestp_isa(Test_env *te, Test_meta meta)
 {
 	int c = tpeek(ARRAYVAR | (meta == TM_BINOP ? 0 : CONTIN));
 	int uqword;
 	char *save = NULL;
-	int ret = 0;
+	Test_op ret = TO_NONOP;
 
 	/* unquoted word? */
 	uqword = c == LWORD && *ident;
 
 	if (meta == TM_OR)
-		ret = c == LOGOR;
+		ret = c == LOGOR ? TO_NONNULL : TO_NONOP;
 	else if (meta == TM_AND)
-		ret = c == LOGAND;
+		ret = c == LOGAND ? TO_NONNULL : TO_NONOP;
 	else if (meta == TM_NOT)
-		ret = uqword && strcmp(yylval.cp, dbtest_tokens[(int) TM_NOT]) == 0;
+		ret = (uqword && !strcmp(yylval.cp,
+		    dbtest_tokens[(int)TM_NOT])) ? TO_NONNULL : TO_NONOP;
 	else if (meta == TM_OPAREN)
-		ret = c == '(' /*)*/;
+		ret = c == '(' /*)*/ ? TO_NONNULL : TO_NONOP;
 	else if (meta == TM_CPAREN)
-		ret = c == /*(*/ ')';
+		ret = c == /*(*/ ')' ? TO_NONNULL : TO_NONOP;
 	else if (meta == TM_UNOP || meta == TM_BINOP) {
 		if (meta == TM_BINOP && c == REDIR &&
 		    (yylval.iop->flag == IOREAD || yylval.iop->flag == IOWRITE)) {
-			ret = 1;
+			ret = TO_NONNULL;
 			save = wdcopy(yylval.iop->flag == IOREAD ?
 			    db_lthan : db_gthan, ATEMP);
 		} else if (uqword && (ret = test_isop(meta, ident)))
 			save = yylval.cp;
 	} else /* meta == TM_END */
-		ret = uqword && strcmp(yylval.cp, db_close) == 0;
-	if (ret) {
+		ret = (uqword && !strcmp(yylval.cp,
+		    db_close)) ? TO_NONNULL : TO_NONOP;
+	if (ret != TO_NONOP) {
 		ACCEPT;
 		if (meta < NELEM(dbtest_tokens))
 			save = wdcopy(dbtest_tokens[(int)meta], ATEMP);
 		if (save)
 			XPput(*te->pos.av, save);
 	}
-	return ret;
+	return (ret);
 }
 
 static const char *
-dbtestp_getopnd(Test_env *te, Test_op op __unused, bool do_eval __unused)
+dbtestp_getopnd(Test_env *te, Test_op op MKSH_A_UNUSED,
+    bool do_eval MKSH_A_UNUSED)
 {
 	int c = tpeek(ARRAYVAR);
 
 	if (c != LWORD)
-		return NULL;
+		return (NULL);
 
 	ACCEPT;
 	XPput(*te->pos.av, yylval.cp);
 
-	return null;
+	return (null);
 }
 
 static int
-dbtestp_eval(Test_env *te __unused, Test_op op __unused,
-    const char *opnd1 __unused, const char *opnd2 __unused,
-    bool do_eval __unused)
+dbtestp_eval(Test_env *te MKSH_A_UNUSED, Test_op op MKSH_A_UNUSED,
+    const char *opnd1 MKSH_A_UNUSED, const char *opnd2 MKSH_A_UNUSED,
+    bool do_eval MKSH_A_UNUSED)
 {
-	return 1;
+	return (1);
 }
 
 static void
